import sys

from .dataload import load_specs
from .smartapi_parser import SmartAPIParser
from .filter import filterOps
import traceback


class MetaKG:
    def __init__(self):
        self.ops = []
        self.parser = SmartAPIParser()

    def populateOpsFromSpecs(self, specs, verbose=False):
        """Populate meta-kg operations based on SmartAPI specifications"""
        self.ops = []
        for spec in specs:
            try:
                self.parser.load_spec(spec)
                self.ops += self.parser.fetch_endpoint_info()
            except:
                if verbose:
                    print("Unexpected error:", sys.exc_info()[0])
                    print(
                        "failed to load the following spec {}".format(spec.get("info"))
                    )

    def constructMetaKG(self, source="remote", tag="translator"):
        """Construct API Meta Knowledge Graph based on SmartAPI Specifications."""
        specs = load_specs(source=source, tag=tag)
        self.populateOpsFromSpecs(specs)

    def filter(self, criteria):
        return filterOps(self.ops, criteria)



import hashlib
from copy import deepcopy

from ..utils.dataload import load_json_or_yaml


class SmartAPIParser:
    """Parse SmartAPI specifications."""

    def load_spec(self, spec=None):
        self.spec = load_json_or_yaml(spec)

    @staticmethod
    def determine_api_type(tags):
        """Determine the type of API
        
        note: API type will be used in the api_preprocess module
        """
        types = ["biothings", "biolink", "reasoner", "ctd", "opentarget"]
        for _type in types:
            if _type in tags:
                return _type
        return "other"

    def fetch_api_name(self):
        """Fetch the name of the API.
        """
        return self.spec["info"]["title"]

    def fetch_api_tags(self):
        """Fetch the tags of the API.
        """
        tags = self.spec["tags"]
        return [_item["name"] for _item in tags]

    def fetch_server_url(self):
        """fetch the server url of the API.
        """
        return self.spec["servers"][0]["url"]

    @staticmethod
    def fetch_path_params(endpoint_spec: dict):
        params = []
        if "parameters" in endpoint_spec:
            for param in endpoint_spec["parameters"]:
                if param.get("in") == "path":
                    params.append(param.get("name"))
        return params

    @staticmethod
    def fetch_response_mapping_path(endpoint_spec):
        """Fetch the path of semantic mapping json doc.
        """
        return endpoint_spec["x-bte-semanticmapping"]["$ref"]

    def fetch_response_mapping(self, ref):
        """Fetch the semantic mapping json file based on the path in $ref.
        """
        path = ref.split("/")
        return self.spec["components"]["x-bte-response-mapping"][path[-1]]

    @staticmethod
    def fetch_param(endpoint_spec):
        """fetch the parameter name of the endpoint

        params
        ======
        endpoint_spec: the smartAPI spec related to the endpoint
        """
        if "parameters" in endpoint_spec:
            return endpoint_spec["parameters"][0].get("name")
        return None

    def fetch_single_x_bte_kgs_operation(self, ref):
        path = ref.split("/")
        return self.spec["components"]["x-bte-kgs-operations"][path[-1]]

    def fetch_x_bte_kgs_operations(self, endpoint_spec):
        """fetch the x-bte-kgs-operations information of the endpoint

        params
        ======
        endpoint_spec: the smartAPI spec related to the endpoint
        """
        operations = endpoint_spec["x-bte-kgs-operations"]
        for op in operations:
            yield self.fetch_single_x_bte_kgs_operation(op["$ref"])

    @staticmethod
    def get_unique_edge_id(operation):
        _id = "-".join(
            [
                operation["query_operation"]["server"],
                operation["query_operation"]["method"],
                operation["query_operation"]["path"],
                operation["association"]["output_type"],
                operation["association"]["output_id"],
            ]
        )
        request_body = operation["query_operation"].get("request_body")
        if request_body and request_body.get("body"):
            for k in sorted(request_body.get("body").keys()):
                _id += "-" + k + "-" + str(request_body.get("body")[k])
        parameters = operation.get("parameters")
        if parameters:
            for k in sorted(parameters.keys()):
                _id += "-" + k + "-" + str(parameters[k])
        return hashlib.sha224(_id.encode()).hexdigest()

    def parse_individual_operation(self, op, path, method, path_params):
        res = []
        for _input in op["inputs"]:
            for _output in op["outputs"]:
                tmp = deepcopy(op)
                tags = self.fetch_api_tags()
                tmp.update(
                    {
                        "query_operation": {
                            "server": self.fetch_server_url(),
                            "params": op.get("parameters"),
                            "request_body": op.get("requestBody"),
                            "path": path,
                            "path_params": path_params,
                            "method": method,
                            "tags": tags,
                            "supportBatch": op.get("supportBatch"),
                            "inputSeparator": op.get("inputSeparator"),
                        },
                        "association": {
                            "input_id": _input["id"],
                            "input_type": _input["semantic"],
                            "output_id": _output["id"],
                            "output_type": _output["semantic"],
                            "predicate": op.get("predicate"),
                            "source": op.get("source"),
                            "api_name": self.fetch_api_name(),
                            "smartapi": {
                                "id": self.spec.get("_id"),
                                "meta": self.spec.get("_meta"),
                            },
                        },
                        "response_mapping": {
                            op["predicate"]: self.fetch_response_mapping(
                                op["response_mapping"]["$ref"]
                            )
                            if "response_mapping" in op
                            else {}
                        },
                    }
                )
                tmp["id"] = self.get_unique_edge_id(tmp)
                res.append(tmp)
        return res

    def fetch_endpoint_info(self):
        res = []
        for path, path_info in self.spec["paths"].items():
            for method, method_info in path_info.items():
                path_params = self.fetch_path_params(method_info)
                if "x-bte-kgs-operations" in method_info:
                    for op in self.fetch_x_bte_kgs_operations(method_info):
                        for _op in op:
                            tmp = self.parse_individual_operation(
                                _op, path, method, path_params
                            )
                            res += tmp
        return res



from .base_transformer import BaseTransformer


class BioLinkTransformer(BaseTransformer):
    def wrap(self, res):
        """ restructure API response from biolink API before extracting data

        parameters
        ----------
        json_doc: the API response from biolink API

        notes: list of prefixes used in biolink API for different semantic types

            * ANATOMY: UBERON, CL, FBbt
            * DISEASE: MONDO
            * GENE: HGNC, NCBIGene, MGI， ZFIN，FlyBase
            * PHENOTYPE: EFO, HP, MONDO
        """
        if res and "associations" in res:
            for _doc in res["associations"]:
                # remove prefix
                if "object" in _doc and "id" in _doc["object"]:
                    object_id = _doc["object"]["id"]
                    try:
                        prefix, value = object_id.split(":")
                        # these IDs have prefix by nature
                        if prefix in ["HGNC", "NCBIGene", "REACT", "dbSNP"]:
                            _doc["object"][prefix] = value
                        else:
                            _doc["object"][prefix] = object_id
                    except ValueError:
                        pass
                # remove empty value
                if "publications" not in _doc:
                    continue
                if not _doc.get("publications"):
                    _doc.pop("publications")
                elif not _doc["publications"][0]["id"].startswith("PMID"):
                    _doc.pop("publications")
                else:
                    for _item in _doc["publications"]:
                        if _item["id"].startswith("PMID"):
                            _item["id"] = _item["id"].split(":")[-1]
                        else:
                            _item.pop("id")
                if not _doc["provided_by"]:
                    _doc.pop("provided_by")
                else:
                    if not isinstance(_doc["provided_by"], list):
                        _doc["provided_by"] = [_doc["provided_by"]]
                    source_list = []
                    for item in _doc["provided_by"]:
                        if item.startswith("https://data.monarchinitiative.org/ttl"):
                            source_name = item.split("/")[-1].strip(".nt")
                            source_list.append(source_name)
                        else:
                            source_name = item.split("/")[-1].strip("#")
                            source_list.append(source_name)
                    _doc["provided_by"] = source_list
        return res



from .base_transformer import BaseTransformer


class CTDTransformer(BaseTransformer):
    def wrap(self, res):
        if isinstance(res, list):
            res = {"data": res}
        for _doc in res["data"]:
            if "PubMedIds" in _doc:
                _doc["PubMedIds"] = _doc["PubMedIds"].split("|")
            if "DiseaseID" in _doc:
                _doc["DiseaseID"] = _doc["DiseaseID"].split(":")[-1]
        return res



from collections import Counter, defaultdict
import requests
import json


class DrugPhaseFilter:
    def __init__(self, stepResult, criteria):
        self.stepResult = stepResult
        self.cnt = Counter()
        self.criteria = criteria

    @staticmethod
    def _extractInputID(resolvedIDs, semanticType):
        if semanticType != "ChemicalSubstance":
            return
        if "db_ids" not in resolvedIDs or not isinstance(resolvedIDs["db_ids"], dict):
            return
        prefix = "CHEMBL.COMPOUND"
        if not prefix in resolvedIDs["db_ids"]:
            return
        return resolvedIDs["db_ids"][prefix]

    @staticmethod
    def _queryDrug(inputs):
        if not inputs:
            return
        result = []
        for i in range(0, len(inputs), 1000):
            a = {
                "q": ",".join(inputs[i : i + 1000]),
                "scopes": "chembl.molecule_chembl_id",
                "fields": "chembl.max_phase",
                "dotfield": True,
            }
            res = requests.post("https://mychem.info/v1/query", json=a,)
            result += res.json()
        return result

    @staticmethod
    def _parseResponse(res):
        if not res:
            return
        result = {}
        for rec in res:
            if "chembl.max_phase" in rec:
                if isinstance(rec["chembl.max_phase"], list):
                    rec["chembl.max_phase"] = max(rec["chembl.max_phase"])
                result[rec["query"]] = rec["chembl.max_phase"]
        return result

    def annotate(self):
        """
        Add node degree info to each edge.
        :param stepResult: a list of returned result from query.
        """
        id_dict = defaultdict(list)
        chembl_ids = set()
        if not self.stepResult:
            return
        for i, rec in enumerate(self.stepResult):
            if "$association" not in rec:
                continue
            output_type = rec["$association"]["output_type"]
            if output_type != "ChemicalSubstance":
                continue
            output_resolved_ids = rec["$output_id_mapping"]["resolved_ids"]
            output_id = self._extractInputID(output_resolved_ids, output_type)
            if not output_id:
                continue
            for o_id in output_id:
                chembl_ids.add(o_id)
                id_dict[i].append(o_id)
        annotatedResult = self._parseResponse(self._queryDrug(list(chembl_ids)))
        cnt = 0
        for i, rec in enumerate(self.stepResult):
            if i in id_dict:
                drug_phases = []
                for o_id in id_dict[i]:
                    if o_id in annotatedResult:
                        drug_phases.append(annotatedResult[o_id])
                if drug_phases:
                    rec["$drug_phase"] = max(drug_phases)
                    cnt += 1
            else:
                rec["$drug_phase"] = 0
        if cnt > 0:
            print(
                "Number of output edges sent to Drug Phase annotation is {}. Number of output edges annotated with Drug Phase is {}".format(
                    len(self.stepResult), cnt
                )
            )
        return

    def filter(self):
        res = []
        if "sort" in self.criteria:
            limit = (
                int(self.criteria["sort"].get("limit"))
                if self.criteria["sort"].get("limit")
                else 0
            )
            if self.criteria["sort"].get("direction") != "descend":
                idList = [item[0] for item in self.cnt.most_common(limit)]
                for rec in self.stepResult:
                    if rec["$output"] in idList:
                        res.append(rec)
                return res
            sortedList = self.cnt.most_common(len(self.cnt))
            sortedList.reverse()
            idList = [item[0] for i, item in enumerate(sortedList) if i < limit]
            for rec in self.stepResult:
                if rec["$output"] in idList:
                    res.append(rec)
            return res



# -*- coding: utf-8 -*-
"""
Storing metadata information and connectivity of APIs.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""
from copy import deepcopy
import networkx as nx
from ._deprecated_mapping_parser import MappingParser
from .config import metadata
from .utils.simple_semmed import semmed
from .utils.cord import cord, SEMANTIC_TYPE_ID_MAPPING
from pathlib import Path
CURRENT_PATH = Path(__file__)

class Registry():

    """Convert metadata information of APIs into a networkx graph."""

    def __init__(self):
        """Initialize networkx graph and load biothings apis."""
        self.G = nx.MultiDiGraph()
        self.registry = {}
        # self.load_biothings()
        self.all_edges_info = self.G.edges(data=True)
        self.all_labels = {d[-1]['label'] for d in self.all_edges_info}
        self.all_inputs = {d[-1]['input_type'] for d in self.all_edges_info}
        self.all_outputs = {d[-1]['output_type'] for d in self.all_edges_info}

    @staticmethod
    def _auto_generate_cord_mapping(doc_type):
        """Auto-generate schema mapping file for all CORD APIs
        
        :param: doc_type: the document type of the specific cord API
        """
        res = {
            "@context": "http://schema.org",
            "@type": doc_type
        }
        for id_type in SEMANTIC_TYPE_ID_MAPPING[doc_type]:
            res[id_type.lower()] = id_type.lower()
        for pred, output_types in cord[doc_type].items():
            res[pred] = []
            for output_type in output_types:
                tmp = {
                    "@type": output_type,
                    "$source": "CORD",
                    "pmd": "associated_with.pmc",
                }
                for input_id_type in SEMANTIC_TYPE_ID_MAPPING[doc_type]:
                    for output_id_type in SEMANTIC_TYPE_ID_MAPPING[output_type]:
                        tmp[output_id_type.lower()] = "associated_with." + output_id_type.lower()
                    tmp["$input"] = input_id_type.lower()
                    res[pred].append(tmp)
        return res

    @staticmethod
    def _auto_generate_semmed_operation_list(doc_type):
        res = []
        for pred, output_types in semmed[doc_type].items():
            for output_type in output_types:
                _id = '-'.join([doc_type, pred, output_type])
                res.append({'$ref': "#/components/x-bte-kgs-operations/" + _id})
        return res

    @staticmethod
    def _auto_generate_semmed_operation(doc_type):
        x_operation_template = {
            "inputSeparator": ",",
            "inputs": [
                {
                    "id": "UMLS"
                }
            ],
            "method": "post",
            "source": "SEMMED",
            "outputs": [
                {
                    "id": "UMLS"
                }
            ],
            "parameters": {
                "fields": ""
            },
            "path": "/query",
            "requestBody": {
                "body": {
                    "q": "{inputs[0]}",
                    "scopes": "umls"
                }
            },
            "supportBatch": True,
            "response_mapping": {
                "$ref": ""
            }
        }
        res = {}
        for pred, output_types in semmed[doc_type].items():
            for output_type in output_types:
                _id = '-'.join([doc_type, pred, output_type])
                tmp = deepcopy(x_operation_template)
                tmp['parameters']['fields'] = pred
                tmp['inputs'][0]['semantic'] = doc_type
                tmp['outputs'][0]['semantic'] = output_type
                tmp['predicate'] = pred
                tmp['response_mapping']["$ref"] = '#/components/x-bte-response-mapping/' + _id
                res[_id] = [tmp]
        return res


    @staticmethod
    def _auto_generate_semmed_mapping(doc_type):
        """Auto-generate schema mapping file for all SEMMED APIs
        
        :param: doc_type: the document type of the specific semmed API
        """
        result = {}
        for pred, output_types in semmed[doc_type].items():
            for output_type in output_types:
                _id = '-'.join([doc_type, pred, output_type])
                result[_id] = {
                    "umls": pred + '.umls',
                    "pmid": pred + '.pmid',
                }
        return result

    def load_biothings(self):
        """Load biothings API into registry network graph."""
        # load biothings schema
        self.mp = MappingParser()
        # loop through API metadata
        for _api, _info in metadata.items():
            # use the mapping parser module to load relationship of each API
            # into the network
            if _info.get('api_name') == 'CORD API':
                mapping_file = self._auto_generate_cord_mapping(_info.get('doc_type'))
            elif _info.get('api_name') == 'SEMMED API':
                mapping_file = self._auto_generate_semmed_mapping(_info.get('doc_type'))
            elif 'mapping_url' in _info:
                self.registry[_api] = {}
                mapping_file = Path.joinpath(CURRENT_PATH.parent,
                                             'smartapi/schema', _api + '.json')
            else:
                continue
            self.mp.load_mapping(mapping_file, api=_api)
            self.registry[_api] = {
                'mapping': self.mp.mapping,
                'graph': self.mp.connect(),
                'type': self.mp.type
            }
            self.G.add_edges_from(self.registry[_api]['graph'].edges(data=True))
        return self.G

    def filter_edges(self, input_cls=None, output_cls=None, edge_label=None):
        """
        Filter edges based on input, output and label.

        The relationship between bio-entities is represented as a networkx MultiDiGraph \
            in BioThings explorer. This function helps you filter for the relationships of your interest based on input/output/edge info.

        :param: input_cls (str|list|None) : the semantic type(s) of the input.
                   Optional
        :param: output_cls (str|list|None) : the semantic type(s) of the output.
                    Optional
        :param: edge_label (str|list|None) : the relationship between input and output.

        """
        if edge_label:
            if isinstance(edge_label, str):
                edge_label = [edge_label]
        # if no edge label is specified, set it as all labels
        else:
            edge_label = self.all_labels
        # if no input_cls is specified, set it as all input types
        if not input_cls:
            input_cls = self.all_inputs
        # if input_cls is str, convert it to list of one element
        elif isinstance(input_cls, str):
            input_cls = [input_cls]
        # if no output_cls is specified, set it as all output types
        if not output_cls:
            output_cls = self.all_outputs
        elif isinstance(output_cls, str):
            output_cls = [output_cls]
        return [d for u,v,d in self.all_edges_info if d['input_type'] in input_cls and d['output_type'] in output_cls and d['label'] in edge_label]



# -*- coding: utf-8 -*-

"""
biothings_explorer.dispatcher
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains code that biothings_explorer use to communicate to and \
    receive from APIs. It serves as a glue between "apicall" module and "api_output_parser" module.
"""
from .json_transformer import Transformer


class OutputParser():
    def __init__(self, res, mapping, batch_mode=False, api=None, api_type=None):
        self.api = api
        self.api_type = api_type
        self.response = res
        self.mapping = mapping
        self.batch_mode = batch_mode

    def parse_biothings_get_res(self):
        """Parse the API response from biothings API using GET method"""
        if self.response['total'] == 0:
            return None
        new_res = {}
        for _res in self.response['hits']:
            transformed_json = Transformer(_res, self.mapping).transform()
            if isinstance(transformed_json, dict):
                for k, v in transformed_json.items():
                    if k in ["@context", "@type"]:
                        new_res[k] = v
                    else:
                        if k not in new_res:
                            new_res[k] = []
                        if isinstance(v, list):
                            new_res[k] += v
                        else:
                            new_res[k].append(v)
            else:
                continue
            return new_res

    def parse_biothings_post_res(self):
        """Parse the API response from biothings API using POST method"""
        new_res = {}
        for _res in self.response:
            if not isinstance(_res, dict):
                continue
            # handle case where the queried item is not found
            if _res.get('notfound'):
                # check if the item is already in final res
                if _res['query'] in new_res:
                    continue
                new_res[_res['query']] = {}
            else:
                # the semmed and cord API are already structured to conform to biolink model
                # so no json transform is needed
                if self.api[:4] in ['semm', 'cord']:
                    transformed_json = _res
                else:
                    transformed_json = Transformer(_res, self.mapping).transform()
                if _res['query'] not in new_res:
                    new_res[_res['query']] = transformed_json
                else:
                    if isinstance(transformed_json, dict):
                        for k, v in transformed_json.items():
                            if k in ["@context", "@type"]:
                                new_res[_res['query']][k] = v
                            else:
                                if k not in new_res[_res['query']]:
                                    new_res[_res['query']][k] = []
                                if isinstance(v, list):
                                    new_res[_res['query']][k] += v
                                else:
                                    new_res[_res['query']][k].append(v)
        return dict(new_res)

    def parse(self):
        if not self.response:
            return None
        # parse the results from BioThings APIs
        if self.api_type == 'biothings':
            if self.batch_mode:
                return self.parse_biothings_post_res()
            return self.parse_biothings_get_res()
        # parse the results from non-BioThings APIs
        return Transformer(self.response, self.mapping).transform()



# -*- coding: utf-8 -*-
"""
Export the output of BTE to GraphML file type.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""

import copy
import networkx as nx


class GraphmlConverter():

    def load_bte_output(self, G):
        """
        Load bte output in the format of networkx graph into class.

        :param: G : the networkx representation of bte output
        """
        self.G = copy.deepcopy(G)

    def restructure_node_info(self):
        """
        Restructure node info.

        note: graphml doesn't support dict as output
        """
        for _, v in self.G.nodes(data=True):
            if 'equivalent_ids' in v:
                v.update(v['equivalent_ids'])
                del v['equivalent_ids']
        for _, v in self.G.nodes(data=True):
            for m, n in v.items():
                if isinstance(n, list):
                    v[m] = ','.join([str(_item) for _item in n])

    def restructure_edge_info(self):
        """
        Restructure edge info.

        note: graphml doesn't support dict as output
        """
        for _, _, o in self.G.edges(data=True):
            if 'info' in o:
                o.update(o['info'])
                del o['info']
            for m, n in o.items():
                if isinstance(n, list):
                    o[m] = ','.join([str(_item) for _item in n])
                if not n:
                    o[m] = 'null'

    def generate_graphml_output(self, path):
        """
        Return the graphml representation of bte output.

        :param: path (str) : the file path to store graphml file
        """
        self.restructure_edge_info()
        self.restructure_node_info()
        return nx.write_graphml(self.G, path)



from pathlib import Path
CURRENT_PATH = Path(__file__)
# these are keys internally used by BTE in the API response mapping file.
INTERNAL_KEYS = ["@context", "@type", "$input", "$source"]

PREFIX_TO_REMOVE = 'bts'

BIOTHINGS_SCHEMA_URL = 'https://raw.githubusercontent.com/data2health/schemas/biothings/biothings/biothings_curie_kevin.jsonld'

id_ranks = {
  'Gene': ['entrez', 'ensembl', 'symbol', 'umls'],
  'SequenceVariant': ['dbsnp', 'hgvs'],
  'ChemicalSubstance': ['chembl', 'drugbank', 'pubchem', 'chebi', 'mesh', 'smiles', 'name', 'umls'],
  'pathway': ['reactome', 'wikipathways', 'kegg', 'pharmgkb', 'biocarta', 'name'],
  'DiseaseOrPhenotypicFeature': ['mondo', 'doid', 'umls', 'mesh', 'hp', 'name'],
  'PhenotypicFeature': ['hp', 'umls'],
  'mf': ['go'],
  'cc': ['go', 'umls'],
  'bp': ['go', 'umls'],
  'anatomy': ['umls'],
  'phenotype': ['umls']
}

metadata = {
  "mygene.info": {
    "scopes": ['entrezgene', 'symbol', 'name', 'HGNC', 'umls.cui', 'uniprot.Swiss-Prot', "ensembl.gene"],
    "id_ranks": ['entrez', 'symbol', 'hgnc', 'umls', 'uniprot', 'ensembl', 'name'],
    "doc_type": "Gene",
    "api_type": "biothings",
    "method": "post",
    "url": "http://mygene.info/v3/query",
    "mapping_url": Path.joinpath(CURRENT_PATH, 'smartapi/schema/mygene.info.json'),
    "hint": True,
    "fields": {
      'entrezgene': 'entrez',
      'name': 'name',
      'symbol': 'symbol',
      'taxid': 'taxonomy',
      'umls.cui': 'umls',
      'uniprot.Swiss-Prot': 'uniprot',
      'HGNC': 'hgnc',
      "ensembl.gene": 'ensembl'
    },
    'api_name': "MyGene.info API",
    'smart_api_id': "59dce17363dce279d389100834e43648"
  },
  "myvariant.info": {
    "scopes": ['dbsnp.rsid', '_id', 'clinvar.rsid',
               'dbnsfp.rsid', 'clinvar.hgvs.coding',
               'clinvar.hgvs.genomic', 'clinvar.hgvs.protein'],
    "id_ranks": ['dbsnp', 'hgvs'],
    "doc_type": "SequenceVariant",
    "hint": True,
    "api_type": "biothings",
    "url": 'http://myvariant.info/v1/query',
    "mapping_url": Path.joinpath(CURRENT_PATH, 'smartapi/schema/myvariant.info.json'),
    "method": "post",
    "fields": {
      '_id': "hgvs",
      'dbsnp.rsid': 'dbsnp'
    },
    'api_name': "MyVariant.info API",
    'smart_api_id': "09c8782d9f4027712e65b95424adba79"
  },
  "mychem.info": {
    "scopes": ['chembl.molecule_chembl_id', 'drugbank.id', 'chebi.id', 'chebi.xrefs.chembl', 'chembl.smiles', 'pubchem.smiles.canonical', 
               'chebi.smiles','drugcentral.structures.smiles', 'pubchem.cid', 'chembl.pref_name', 'drugbank.name', 
               'unii.unii', 'ginas.preferred_name', 'drugcentral.xrefs.umlscui', "drugcentral.synonyms", 
               "ginas.name_list", "drugcentral.xrefs.chebi", "drugcentral.xrefs.mesh_descriptor_ui"],
    "id_ranks": ['chembl', 'drugbank', 'pubchem', 'mesh', 'chebi', 'name'],
    "doc_type": "ChemicalSubstance",
    "api_type": "biothings",
    "hint": True,
    "url": "http://mychem.info/v1/query",
    "mapping_url": Path.joinpath(CURRENT_PATH, 'smartapi/schema/mychem.info.json'),
    "method": "post",
    "fields": {
      'chembl.molecule_chembl_id': 'chembl',
      'chebi.xrefs.chembl': 'chembl',
      'drugbank.id': 'drugbank',
      'drugbank.name': 'name',
      'chembl.pref_name': 'name',
      'pubchem.cid': 'pubchem',
      'drugcentral.xrefs.umlscui': 'umls',
      'drugcentral.xrefs.mesh_descriptor_ui': 'mesh',
      'chebi.id': 'chebi',
      'chebi.xrefs.pubchem.cid': 'pubchem',
      'drugcentral.xrefs.chebi': 'chebi',
      'chembl.smiles': 'smiles',
      'pubchem.smiles.canonical': 'smiles',
      'chebi.smiles': 'smiles',
      'drugcentral.structures.smiles': 'smiles'
    },
    'api_name': "MyChem.info API",
    "smart_api_id": "8f08d1446e0bb9c2b323713ce83e2bd3"
  },
  "mydisease.info": {
    "scopes": ['_id', 'mondo.xrefs.doid', 'mondo.xrefs.hp',
               'mondo.xrefs.mesh', 'mondo.xrefs.umls',
               'mondo.label', 'disgenet.xrefs.disease_name'],
    "id_ranks": ['mondo', 'doid', 'umls', 'mesh', 'name'],
    "doc_type": "DiseaseOrPhenotypicFeature",
    "api_type": "biothings",
    "hint": True,
    "url": "http://mydisease.info/v1/query",
    "mapping_url": 'https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/mydisease.info/schema.json',
    "method": "post",
    "fields": {
      '_id': "mondo",
      'mondo.xrefs.doid': 'doid',
      'mondo.xrefs.hp': 'hp',
      'mondo.xrefs.umls': 'umls',
      'mondo.xrefs.mesh': 'mesh',
      'mondo.label': 'name',
      'disgenet.xrefs.disease_name': 'name'
    },
    'api_name': "MyDisease.info API",
    "smart_api_id": "f307760715d91908d0ae6de7f0810b22"
  },
  "semmeddisease": {
    "url": "http://biothings.ncats.io/semmed/query",
    "mapping_url": 'https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/semmed/schema.json',
    'api_type': "biothings",
    'api_name': 'SEMMED API',
    'doc_type': 'DiseaseOrPhenotypicFeature',
    "smart_api_id": "e99229fc6ccb9ad9889bcc9c77a36bad"
  },
  "semmedanatomy": {
    "url": "https://biothings.ncats.io/semmed_anatomy/query",
    "mapping_url": 'https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/semmedanatomy/schema.json',
    'api_type': "biothings",
    'api_name': 'SEMMED API',
    'doc_type': 'AnatomicalEntity',
    "smart_api_id": "e99229fc6ccb9ad9889bcc9c77a36bad"
  },
  "semmedbp": {
    "url": "https://biothings.ncats.io/semmedbp/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/semmedbp/schema.json",
    'api_type': "biothings",
    'api_name': 'SEMMED API',
    'doc_type': 'BiologicalProcess',
    "smart_api_id": "e99229fc6ccb9ad9889bcc9c77a36bad"
  },
  "semmedchemical": {
    "url": "https://biothings.ncats.io/semmedchemical/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/semmedchemical/schema.json",
    'api_type': "biothings",
    'api_name': 'SEMMED API',
    'doc_type': 'ChemicalSubstance',
    "smart_api_id": "e99229fc6ccb9ad9889bcc9c77a36bad"
  },
  "semmedgene": {
    "url": "https://biothings.ncats.io/semmedgene/query",
    "mapping_url": 'https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/semmedgene/schema.json',
    'api_type': "biothings",
    'api_name': 'SEMMED API',
    'doc_type': 'Gene',
    "smart_api_id": "e99229fc6ccb9ad9889bcc9c77a36bad"
  },
  "semmedphenotype": {
    "url": "https://biothings.ncats.io/semmedphenotype/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/semmedphenotype/schema.json",
    'api_type': "biothings",
    'api_name': 'SEMMED API',
    'doc_type': 'PhenotypicFeature',
    "smart_api_id": "e99229fc6ccb9ad9889bcc9c77a36bad"
  },
  "cordgene": {
    "url": "https://biothings.ncats.io/cord_gene/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'Gene',
    "smart_api_id": "6bc54230a6fa7693b2cd113430387ca7"
  },
  "cordprotein": {
    "url": "https://biothings.ncats.io/cord_protein/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'Protein',
    "smart_api_id": "1066244f2976e931654394581255630e"
  },
  "cordchemical": {
    "url": "https://biothings.ncats.io/cord_chemical/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'ChemicalSubstance',
    "smart_api_id": "503e8c42d041aa5b8b1ffafaf612c298"
  },
  "cordcell": {
    "url": "https://biothings.ncats.io/cord_cell/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'Cell',
    "smart_api_id": "3d6bacf6e305bee5efdd25d1560996e1"
  },
  "corddisease": {
    "url": "https://biothings.ncats.io/cord_disease/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'DiseaseOrPhenotypicFeature',
    "smart_api_id": "871d5b6761d91a7805572c47f016fe47"
  },
  "cordma": {
    "url": "https://biothings.ncats.io/cord_ma/query",
    'api_type': "biothings",
    'api_name': 'CORD API ',
    'doc_type': 'MolecularActivity',
    "smart_api_id": "066a2ef51223a5a95216355facaac9d0"
  },
  "cordbp": {
    "url": "https://biothings.ncats.io/cord_bp/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'BiologicalProcess',
    "smart_api_id": "7cbc3ff9b2d09cb8075715bde740ce4f"
  },
  "cordgenomicentity": {
    "url": "https://biothings.ncats.io/cord_genomic_entity/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'GenomicEntity',
    "smart_api_id": "bffe279a1387ef956a423e2e8a133cf6"
  },
  "cordanatomy": {
    "url": "https://biothings.ncats.io/cord_anatomy/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'AnatomicalEntity',
    "smart_api_id": "52f5867b9cc9cf8664565cc9ef2af276"
  },
  "cordcc": {
    "url": "https://biothings.ncats.io/cord_cc/query",
    'api_type': "biothings",
    'api_name': 'CORD API',
    'doc_type': 'CellularComponent',
    "smart_api_id": "bf5aac143254714988018d87dc447a5c"
  },
  "mgigene2phenotype": {
    "url": "https://biothings.ncats.io/mgigene2phenotype/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/mgigene2phenotype/schema.json",
    'api_type': "biothings",
    'api_name': 'MGIgene2phenotype API',
    "smart_api_id": "77ed27f111262d0289ed4f4071faa619"
  },
  "ebigene2phenotype": {
    "url": "https://biothings.ncats.io/ebigene2phenotype/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/ebigene2phenotype/schema.json",
    'api_type': "biothings",
    'api_name': 'EBIgene2phenotype API',
    "smart_api_id": "1f47552dabd67351d4c625adb0a10d00"
  },
  "DISEASES": {
    "url": "https://biothings.ncats.io/DISEASES/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/DISEASES/schema.json",
    'api_type': "biothings",
    'api_name': 'DISEASES API',
    'smart_api_id': "a7f784626a426d054885a5f33f17d3f8"
  },
  "pfocr": {
    "url": "https://biothings.ncats.io/pfocr/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/pfocr/schema.json",
    'api_type': "biothings",
    'api_name': 'pfocr API',
    "smart_api_id": "edeb26858bd27d0322af93e7a9e08761"
  },
  "pathway": {
    "scopes": ['_id', 'name'],
    "id_ranks": ['reactome', 'wikipathways', 'kegg', 'pharmgkb', 'biocarta', 'name'],
    "doc_type": "Pathway",
    "api_type": "biothings",
    "method": "get",
    "url": "http://biothings.ncats.io/geneset/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/geneset/pathway_schema.json",
    "add": " AND type:pathway",
    "hint": True,
    "fields": {
      'name': 'name',
      'reactome': 'reactome',
      'wikipathways': 'wikipathways',
      'kegg': 'kegg',
      'pharmgkb': 'pharmgkb',
      'biocarta': 'biocarta'
    }
  },
  "mf": {
    "scopes": ['_id', 'name'],
    "id_ranks": ['go', 'name'],
    "doc_type": "MolecularActivity",
    "api_type": "biothings",
    "add": " AND type:mf",
    "method": "get",
    "hint": True,
    "url": "http://biothings.ncats.io/geneset/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/geneset/mf_schema.json",
    "fields":{
      'name': 'name',
      'go': 'go'
    }
  },
  "cc": {
    "scopes": ['_id', 'name'],
    "id_ranks": ['go', 'umls', 'name'],
    "doc_type": "CellularComponent",
    "api_type": "biothings",
    "method": "get",
    "add": " AND type:cc",
    "hint": True,
    "url": "http://biothings.ncats.io/geneset/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/geneset/cc_schema.json",
    "fields":{
      'name': 'name',
      'go': 'go',
      'umls': 'umls'
    }
  },
  "bp": {
    "scopes": ['_id', 'name'],
    "id_ranks": ['go', 'umls', 'name'],
    "doc_type": "BiologicalProcess",
    "api_type": "biothings",
    "add": " AND type:bp",
    "method": "get",
    "hint": True,
    "url": "http://biothings.ncats.io/geneset/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/geneset/bp_schema.json",
    "fields": {
      'name': 'name',
      'go': 'go',
      'umls': 'umls'
    }
  },
  "anatomy": {
    "scopes": ['umls', 'name'],
    "id_ranks": ['umls', 'name'],
    "doc_type": "Anatomy",
    "api_type": "biothings",
    "hint": True,
    "url": "http://biothings.ncats.io/semmed_anatomy/query",
    "method": "post",
    "fields": {
      "name": "name",
      "umls": "umls"
    }
  },
  "phenotype": {
    "scopes": ['umls', 'name'],
    "id_ranks": ['umls', 'name'],
    "doc_type": "PhenotypicFeature",
    "api_type": "biothings",
    "url": "http://biothings.ncats.io/semmedphenotype/query",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/semmedphenotype/schema.json",
    "method": "post",
    "fields": {
      "name": "name",
      "umls": "umls"
    },
    "hint": True
  },
  "umlschem": {
    "scopes": ['umls', 'name'],
    "id_ranks": ['umls', 'name'],
    "doc_type": "ChemicalSubstance",
    "api_type": "biothings",
    "hint": True,
    "url": "http://biothings.ncats.io/umlschem/query",
    "method": "post",
    "fields": {
      "name": "name",
      "umls": "umls"
    }
  },
  "biolink_anatomy2gene": {
    "url": "https://api.monarchinitiative.org/api/bioentity/anatomy/{anatomy_id}/genes?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_anatomy2gene.json",
    "method": "get",
    "path": "anatomy_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_disease2gene": {
    "url": "https://api.monarchinitiative.org/api/bioentity/disease/{disease_id}/genes?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_disease2gene.json",
    "method": "get",
    "path": "disease_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },

  "biolink_disease2pathway": {
    "url": "https://api.monarchinitiative.org/api/bioentity/disease/{disease_id}/pathways?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_disease2pathway.json",
    "method": "get",
    "path": "disease_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_disease2phenotype": {
    "url": "https://api.monarchinitiative.org/api/bioentity/disease/{disease_id}/phenotypes?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_disease2phenotype.json",
    "method": "get",
    "path": "disease_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_gene2anatomy": {
    "url": "https://api.monarchinitiative.org/api/bioentity/gene/NCBIGene:{gene_id}/anatomy?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_gene2anatomy.json",
    "method": "get",
    "path": "gene_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_gene2disease": {
    "url": "https://api.monarchinitiative.org/api/bioentity/gene/NCBIGene:{gene_id}/diseases?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_gene2disease.json",
    "method": "get",
    "path": "gene_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_geneinteraction": {
    "url": "https://api.monarchinitiative.org/api/bioentity/gene/NCBIGene:{gene_id}/interactions?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_geneinteraction.json",
    "method": "get",
    "path": "gene_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_gene2phenotype": {
    "url": "https://api.monarchinitiative.org/api/bioentity/gene/NCBIGene:{gene_id}/phenotypes?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_gene2phenotype.json",
    "method": "get",
    "path": "gene_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_pathway2disease": {
    "url": "https://api.monarchinitiative.org/api/bioentity/pathway/REACT:{pathway_id}/diseases?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_pathway2disease.json",
    "method": "get",
    "path": "pathway_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_pathway2phenotype": {
    "url": "https://api.monarchinitiative.org/api/bioentity/pathway/REACT:{pathway_id}/phenotypes?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_pathway2phenotype.json",
    "method": "get",
    "path": "pathway_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_phenotype2disease": {
    "url": "https://api.monarchinitiative.org/api/bioentity/phenotype/{phenotype_id}/diseases?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_phenotype2disease.json",
    "method": "get",
    "path": "phenotype_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_phenotype2gene": {
    "url": "https://api.monarchinitiative.org/api/bioentity/phenotype/{phenotype_id}/genes?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_phenotype2gene.json",
    "method": "get",
    "path": "phenotype_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "biolink_phenotype2pathway": {
    "url": "https://api.monarchinitiative.org/api/bioentity/phenotype/{phenotype_id}/pathways?rows=100",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/biolink/schema/biolink_phenotype2pathway.json",
    "method": "get",
    "path": "phenotype_id",
    'api_type': "biolink",
    'api_name': "BioLink API",
    "smart_api_id": "d22b657426375a5295e7da8a303b9893"
  },
  "dgidb_gene2chemical": {
    "url": "http://www.dgidb.org/api/v2/interactions.json?genes={gene_id}",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/dgidb/schema/dgidb_gene2chemical.json",
    "method": "get",
    "path": "gene_id",
    'api_type': "other",
    'api_name': "DGIdb API",
    "smart_api_id": "e3edd325c76f2992a111b43a907a4870"
  },
  "dgidb_chemical2gene": {
    "url": "http://www.dgidb.org/api/v2/interactions.json?drugs={chembl_id}",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/dgidb/schema/dgidb_chemical2gene.json",
    "method": "get",
    "path": "chembl_id",
    'api_type': "other",
    'api_name': "DGIdb API",
    "smart_api_id": "e3edd325c76f2992a111b43a907a4870"
  },
  "quickgo_children": {
    "url": "https://www.ebi.ac.uk/QuickGO/services/ontology/go/terms/{goterm}/children",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/quickgo/schema/children.json",
    "method": "get",
    "path": "goterm",
    'api_type': "other",
    'api_name': 'QuickGO API',
    "smart_api_id": "1f277e1563fcfd124bfae2cc3c4bcdec"
  },
  "litvar": {
    "url": "https://www.ncbi.nlm.nih.gov/research/bionlp/litvar/api/v1/entity/litvar/{dbsnp}##",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/litvar/schema/schema.json",
    "method": "get",
    "path": "dbsnp",
    'api_type': "other",
    'api_name': 'LitVar API',
    "smart_api_id": "dca415f2d792976af9d642b7e73f7a41"
  },
  "ols_children": {
    "url": "https://www.ebi.ac.uk/ols/api/ontologies/doid/children?id={doid}",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/ols/schema/schema.json",
    "method": "get",
    "path": "doid",
    'api_type': "other",
    'api_name': "Ontology Lookup Service API",
    "smart_api_id": "1c056ffc7ed0dd1229e71c4752239465"
  },
  "stanford_biosample_disease2sample": {
    "url": "http://api.kp.metadatacenter.org/biosample/search?q=biolink:Disease={mondo}&limit=1000",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/stanford/disease2sample.json",
    "method": "get",
    "path": "mondo",
    "api_type": "stanford",
    "api_name": "Stanford BioSample API",
    "smart_api_id": "553a49d112bb19306253942ebd6377a9"
  },
  "stanford_biosample_cl2sample": {
    "url": "http://api.kp.metadatacenter.org/biosample/search?q=biolink:CellLine={clo}&limit=1000",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/stanford/cl2sample.json",
    "method": "get",
    "path": "clo",
    "api_type": "stanford",
    "api_name": "Stanford BioSample API",
    "smart_api_id": "553a49d112bb19306253942ebd6377a9"
  },
  "chembl_drug_mechanism": {
    "url": "https://www.ebi.ac.uk/chembl/api/data/mechanism.json?molecule_chembl_id={chembl}",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/ChEMBL/drugmechanism.json",
    "method": "get",
    "path": "chembl",
    "api_type": "other",
    "api_name": "ChEMBL API",
    "smart_api_id": "71add13e7c8b26b0046cfb8cf5092395"
  },
  "ctd_chemical2gene": {
    "url": "http://ctdbase.org/tools/batchQuery.go?inputType=chem&inputTerms={mesh}|mercury&report=genes_curated&format=json",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/CTD_api/chemical2gene.json",
    "method": "get",
    "path": "mesh",
    "api_type": "ctd",
    "api_name": "CTD API",
    "smart_api_id": "0212611d1c670f9107baf00b77f0889a"
  },
  "ctd_gene2disease": {
    "url": "http://ctdbase.org/tools/batchQuery.go?inputType=gene&inputTerms={entrez}&report=diseases_curated&format=json",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/CTD_api/gene2disease.json",
    "method": "get",
    "path": "entrez",
    "api_type": "ctd",
    "api_name": "CTD API",
    "smart_api_id": "0212611d1c670f9107baf00b77f0889a"
  },
  "opentarget": {
    "url": "https://platform-api.opentargets.io/v3/platform/public/evidence/filter?target={ensembl}&datasource=chembl&size=15&fields=drug",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/opentarget/gene2chemical.json",
    "method": "get",
    "path": "ensembl",
    "api_type": "opentarget",
    "api_name": "OpenTarget API",
    "smart_api_id": "542ba165e1b4227854cf7c0a8addcc79"
  },
  "RGD": {
    "url": "https://rest.rgd.mcw.edu/rgdws/genes/{rgd}",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/RGD/schema.json",
    "method": "get",
    "path": "rgd",
    "api_type": "other",
    "api_name": "RGD API",
    "smart_api_id": "eb28713e2e23160e80e88f3a5dabcca4"
  },
  "scibite_gene2chemical": {
    "url": "https://automat.renci.org/cord19_scibite_v2/gene/chemical_substance/NCBIGene:{entrez}",
    "path": "entrez",
    "method": "get",
    "api_type": "automat",
    "api_name": "Automat CORD19 Scibite API",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/scibite_gene2chemical/schema.json",
    "smart_api_id": "5c5a0b5bc77e5d25d1a04e1385f9fad7"
  },
  "scibite_chemical2gene": {
    "url": "https://automat.renci.org/cord19_scibite_v2/chemical_substance/gene/{chebi}",
    "path": "chebi",
    "method": "get",
    "api_type": "automat",
    "api_name": "Automat CORD19 Scibite API",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/scibite_chemical2gene/schema.json",
    "smart_api_id": "5c5a0b5bc77e5d25d1a04e1385f9fad7"
  },
  "scibite_chemical2disease": {
    "url": "https://automat.renci.org/cord19_scibite_v2/chemical_substance/disease/{chebi}",
    "path": "chebi",
    "method": "get",
    "api_type": "automat",
    "api_name": "Automat CORD19 Scibite API",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/scibite_chemical2disease/schema.json",
    "smart_api_id": "5c5a0b5bc77e5d25d1a04e1385f9fad7"
  },
  "scibite_gene2disease": {
    "url": "https://automat.renci.org/cord19_scibite_v2/gene/disease/NCBIGene:{entrez}",
    "path": "entrez",
    "method": "get",
    "api_type": "automat",
    "api_name": "Automat CORD19 Scibite API",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/scibite_gene2disease/schema.json",
    "smart_api_id": "5c5a0b5bc77e5d25d1a04e1385f9fad7"
  },
  "scibite_disease2gene": {
    "url": "https://automat.renci.org/cord19_scibite_v2/disease/gene/{mondo}",
    "path": "mondo",
    "method": "get",
    "api_type": "automat",
    "api_name": "Automat CORD19 Scibite API",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/scibite_disease2gene/schema.json",
    "smart_api_id": "5c5a0b5bc77e5d25d1a04e1385f9fad7"
  },
  "scibite_disease2chemical": {
    "url": "https://automat.renci.org/cord19_scibite_v2/disease/chemical_substance/{mondo}",
    "path": "mondo",
    "method": "get",
    "api_type": "automat",
    "api_name": "Automat CORD19 Scibite API",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/scibite_disease2chemical/schema.json",
    "smart_api_id": "5c5a0b5bc77e5d25d1a04e1385f9fad7"
  }
}


METADATA_OPTIOINAL = {
  "robokop_gene2chemical": {
    "url": "https://robokop.renci.org/api/simple/expand/gene/HGNC:{hgnc_id}/chemical_substance/",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/robokop/gene2chemical.json",
    "method": "get",
    "path": "hgnc_id",
    'api_type': "reasoner",
    'api_name': "robokop"
  },
  "robokop_gene2genefamily": {
    "url": "https://robokop.renci.org/api/simple/expand/gene/HGNC:{hgnc_id}/gene_family/",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/robokop/gene2genefamily.json",
    "method": "get",
    "path": "hgnc_id",
    'api_type': "reasoner",
    'api_name': "robokop"
  },
  "robokop_chemical2disease": {
    "url": "https://robokop.renci.org/api/simple/expand/chemical_substance/CHEMBL:{hgnc_id}/disease_or_phenotypic_feature/",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/robokop/chemical2disease.json",
    "method": "get",
    "path": "hgnc_id",
    'api_type': "reasoner",
    'api_name': "robokop"
  },
  "cohd_disease2chemical": {
    "url": "http://cohd.io/api/frequencies/associatedConceptDomainFreq?dataset_id=1&concept_id={cohd}&domain=Drug",
    "mapping_url": "https://raw.githubusercontent.com/NCATS-Tangerine/translator-api-registry/openapi_2.0/cohd/schema/disease2chemical.json",
    "method": "get",
    "path": "cohd",
    'api_type': "other",
    'api_name': 'cohd'
  },
}


# -*- coding: utf-8 -*-
"""
Export the output of BTE to pandas data frame.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""

import pandas as pd


def networkx2pandas(current_graph, input_type):
    """Converting current graph into a pandas data frame

    :param: current_graph: a python dict containing all paths
    :param: input_type: the semantic type of the input
    """
    data = []
    for paths in current_graph.values():
        for path in paths:
            path_data = {}
            for i, edge in enumerate(path):
                if i == 0:
                    path_data['input'] = edge['input_name']
                    path_data['input_type'] = input_type
                path_data['pred' + str(i + 1)] = edge['info']['label']
                path_data['pred' + str(i + 1) +
                          '_source'] = retrieve_prop_from_edge(edge['info'],
                                                               'source')
                path_data['pred' + str(i + 1) +
                          '_api'] = retrieve_prop_from_edge(edge['info'],
                                                            'api')
                path_data['pred' + str(i + 1) +
                          '_pubmed'] = retrieve_prop_from_edge(edge['info'],
                                                               'pubmed')
                node = 'output' if i + 1 == len(path) else 'node' + str(i + 1)
                path_data[node + '_type'] = edge['info']['info']['@type']
                path_data[node + '_name'] = edge['output_name']
                path_data[node + '_id'] = edge['output_id']
            data.append(path_data)
    return pd.DataFrame(data).drop_duplicates()


def retrieve_prop_from_edge(edge_info, prop):
    """Retrieve property info from the edge data."""
    if prop == 'api':
        data = edge_info['info'].get('$api')
        if data:
            if not isinstance(data, list):
                data = [data]
            return ','.join(data)
    elif prop == 'source':
        data = edge_info['info'].get('$source')
        if data:
            if not isinstance(data, list):
                data = [data]
            return ','.join(data)
    elif prop == 'pubmed':
        data = edge_info['info'].get('pubmed')
        if data:
            if not isinstance(data, list):
                data = [data]
            else:
                data = [str(item) if not isinstance(item, str) else item
                        for item in data]
            return ','.join(data)



from collections import defaultdict
from copy import deepcopy
import pandas as pd

from ..config_new import ALWAYS_PREFIXED
from ..smartapi_kg import MetaKG


def id2curie(prefix, val):
    if prefix in ALWAYS_PREFIXED:
        return val
    return prefix + ":" + val


def annotateEdgesWithInput(edges, inputs):
    if isinstance(inputs, dict):
        inputs = [inputs]
    annotatedEdges = []
    for edge in edges:
        if edge["query_operation"].get("supportBatch"):
            copy_edge = deepcopy(edge)
            input_ids = set()
            original_input = {}
            for _input in inputs:
                prefix = copy_edge["association"]["input_id"]
                if prefix in _input["db_ids"]:
                    for val in _input["db_ids"][prefix]:
                        input_ids.add(val)
                        original_input[id2curie(prefix, val)] = _input
            input_ids = list(input_ids)
            step = 500
            for i in range(0, len(input_ids), step):
                copy_edge["input"] = input_ids[i : i + step]
                copy_edge["original_input"] = original_input
                annotatedEdges.append(deepcopy(copy_edge))
        else:
            for _input in inputs:
                prefix = edge["association"]["input_id"]
                if prefix in _input["db_ids"]:
                    if not isinstance(_input["db_ids"][prefix], list):
                        _input["db_ids"][prefix] = [_input["db_ids"][prefix]]
                    for _id in _input["db_ids"][prefix]:
                        copy_edge = deepcopy(edge)
                        copy_edge["input"] = _id
                        copy_edge["original_input"] = {id2curie(prefix, _id): _input}
                        annotatedEdges.append(deepcopy(copy_edge))
    return annotatedEdges


def getEdges(inputs, outputs, predicates, knowledgegraph=None):
    result = []
    if not knowledgegraph:
        kg = MetaKG()
        kg.constructMetaKG(source="local")
    else:
        kg = knowledgegraph
    for semantic_type, ids in inputs.items():
        if (
            (isinstance(outputs, list) and len(outputs) == 1)
            or not isinstance(outputs, list)
            or not isinstance(predicates, list)
        ):
            edges = kg.filter(
                {
                    "input_type": semantic_type,
                    "output_type": outputs,
                    "predicate": predicates,
                }
            )
        else:
            edges = []
            if not isinstance(predicates, list):
                predicates = [predicates]
            for i, node in enumerate(outputs):
                if i >= len(predicates):
                    tmp_predicate = None
                else:
                    tmp_predicate = predicates[i]
                edges += kg.filter(
                    {
                        "input_type": semantic_type,
                        "output_type": node,
                        "predicate": tmp_predicate,
                    }
                )
        if not edges or not ids:
            continue
        result.append({"edges": edges, "inputs": ids})
    return result


def extractAllResolvedOutputIDs(res):
    output_ids = {}
    if res and len(res) > 0:
        for item in res:
            if "resolved_ids" in item["$output_id_mapping"]:
                output_ids[
                    item["$output_id_mapping"]["resolved_ids"]["id"]["identifier"]
                ] = item["$output_id_mapping"]["resolved_ids"]
    return output_ids


def groupsIDsbySemanticType(output_ids):
    result = defaultdict(list)
    for resolved_ids in output_ids.values():
        result[resolved_ids.get("type")].append(resolved_ids)
    return result


def restructureHintOutput(outputs):
    result = {}
    for output in outputs:
        copy_output = deepcopy(output)
        output_id = {}
        for k, v in copy_output.items():
            if k not in ["primary", "type"] and not isinstance(v, list):
                copy_output[k] = [v]
        if copy_output["primary"]["identifier"] in ALWAYS_PREFIXED:
            curie = copy_output["primary"]["value"]
        else:
            curie = (
                copy_output["primary"]["identifier"]
                + ":"
                + copy_output["primary"]["value"]
            )
        if "name" in copy_output:
            output_id["label"] = copy_output["name"][0]
        else:
            output_id["label"] = curie
        output_id["identifier"] = curie
        copy_output.pop("display")
        copy_output.pop("primary")
        result.update(
            {
                curie: {
                    "type": copy_output.pop("type"),
                    "db_ids": copy_output,
                    "id": output_id,
                }
            }
        )
    return result


def stepResult2PandasTable(result, step, total_steps, extra_fields=[]):
    if step == 0:
        node1 = "input"
    else:
        node1 = "node" + str(step)
    if step == total_steps - 1:
        node2 = "output"
    else:
        node2 = "node" + str(step + 1)
    if isinstance(result, list) and len(result) > 1:
        table_dict = []
        for rec in result:
            d = {
                node1
                + "_id": rec["$original_input"][rec["$input"]]["id"]["identifier"],
                node1 + "_label": rec["$original_input"][rec["$input"]]["id"]["label"],
                node1 + "_type": rec["$original_input"][rec["$input"]]["type"],
                "pred" + str(step + 1): rec["$association"]["predicate"],
                "pred" + str(step + 1) + "_source": ",".join(rec.get("provided_by"))
                if rec.get("provided_by") != [None]
                else None,
                "pred" + str(step + 1) + "_api": rec.get("api"),
                "pred"
                + str(step + 1)
                + "_publications": ",".join(rec.get("publications"))
                if rec.get("publications")
                else None,
                node2
                + "_id": rec["$output_id_mapping"]["resolved_ids"]["id"]["identifier"],
                node2
                + "_label": rec["$output_id_mapping"]["resolved_ids"]["id"]["label"],
                node2 + "_type": rec["$output_id_mapping"]["resolved_ids"]["type"],
                node2 + "_degree": rec.get("$nodeDegree"),
            }
            if isinstance(extra_fields, list) and len(extra_fields) > 0:
                for field in extra_fields:
                    if field in [
                        "drug_phase",
                        "ngd",
                        "survival_prob_change",
                        "edgesOut",
                    ]:
                        field = "$" + field
                    if field in rec:
                        if field in ["pvalue"]:
                            rec[field] = float(rec[field])
                        elif isinstance(rec[field], list) and len(rec[field]) == 1:
                            rec[field] = rec[field][0]
                        d.update(
                            {
                                "pred"
                                + str(step + 1)
                                + "_"
                                + str(field).strip("$"): rec[field]
                            }
                        )
            table_dict.append(d)

        return pd.DataFrame(table_dict)


def validate_max_intermediate_nodes(intermediate_nodes):
    """
    Validate if user inputs more than 2 intermediate nodes.
    :param intermediate_nodes: a list of intermediate nodes
    """
    if isinstance(intermediate_nodes, list) and len(intermediate_nodes) > 3:
        print(
            "Max number of intermediate nodes is 2. You specify {}. We can not proceed your query. Please refine your query".format(
                len(intermediate_nodes) - 1
            )
        )
        return False
    return True



from copy import deepcopy


def restructure_semmed_response(json_doc, output_types):
    """Restructure the JSON output from semmed API.

    :param: json_doc: the API response from semmed API
    """
    if not isinstance(json_doc, list):
        return json_doc
    new_res = []
    for _res in json_doc:
        if not isinstance(_res, dict):
            continue
        # handle case where the queried item is not found
        if _res.get('notfound'):
            continue
        else:
            tmp_res = {'query': _res['query']}
            for k, v in _res.items():
                tmp_v = []
                if isinstance(v, list):
                    for item in v:
                        if item["@type"] in output_types or (item["@type"] == 'DiseaseOrPhenotypicFeature' and output_types == ['Disease']):
                            item_copy = deepcopy(item)
                            item_copy['UMLS'] = item_copy.pop('umls')
                            item_copy['pubmed'] = item_copy.pop('pmid')
                            tmp_v.append(item_copy)
                if tmp_v != []:
                    tmp_res[k] = tmp_v
        new_res.append(tmp_res)
    return new_res



from ..config_new import ID_RESOLVING_APIS
from ..utils.common import getCurieFromVal


class BioEntity:
    def __init__(self, semantic_type, id_dict):
        self._semantic_type = semantic_type
        self._id_dict = id_dict

    def __repr__(self):
        return self._id_dict

    def get_primary_id(self):
        ranks = ID_RESOLVING_APIS.get(self._semantic_type, {}).get("id_ranks")
        for prefix in ranks:
            if self._id_dict.get(prefix):
                return getCurieFromVal(self._id_dict.get(prefix)[0], prefix)
        return None

    def get_label(self):
        if self._id_dict.get("SYMBOL"):
            return self._id_dict["SYMBOL"][0]
        if self._id_dict.get("name"):
            return self._id_dict["name"][0]
        return self.get_primary_id()

    def get_curies(self):
        res = []
        for prefix, ids in self._id_dict.items():
            for _id in ids:
                res.append(getCurieFromVal(_id, prefix))
        return res



from collections import defaultdict


def restructure_scibite_response(json_doc):
    """restructure the JSON output from AUTOMAT SCIBITE API

    parameters
        * json_doc: the API response from AUTOMAT SCIBITE API
    """
    # convert from list to dict
    if json_doc['data'] == []:
        return {}
    res = defaultdict(set)
    for rec in json_doc['data']:
        obj = rec[-1]
        obj_id_prefix = obj['id'].split(':')[0]
        if obj_id_prefix in ['MONDO', 'CHEBI', 'DOID']:
            res[obj_id_prefix.lower()].add(obj['id'])
        elif obj_id_prefix in ['NCBIGene', 'ENSEMBL']:
            res[obj_id_prefix.lower()].add(obj['id'].split(':', 1)[-1])
        elif obj_id_prefix == 'CHEMBL.COMPOUND':
            res['chembl'].add(obj['id'].split(':', 1)[-1])
        obj = rec[0]
        obj_id_prefix = obj['id'].split(':')[0]
        if obj_id_prefix in ['MONDO', 'CHEBI', 'DOID']:
            res[obj_id_prefix.lower()].add(obj['id'])
        elif obj_id_prefix in ['NCBIGene', 'ENSEMBL']:
            res[obj_id_prefix.lower()].add(obj['id'].split(':', 1)[-1])
        elif obj_id_prefix == 'CHEMBL.COMPOUND':
            res['chembl'].add(obj['id'].split(':', 1)[-1])
    new_res = {}
    for k, v in res.items():
        new_res[k] = list(v)
    return {'associated_with': new_res}



from functools import partial

# from ..config_new import BTE_FILTERS
BTE_FILTERS = ["nodeDegree", "ngd", "drugPhase", "survivalProbability"]


def filter_response(res, criteria):
    """
    Filter API response based on filtering criteria
    :param res: API Response
    :param criteria: filtering criteria
    """

    def filter_by_operation(rec, key, val, operation):
        if rec.get(key):
            if isinstance(rec.get(key), list):
                rec[key] = rec[key][0]
            try:
                if operation == "=" and type(val)(rec[key]) == val:
                    return True
                if operation == ">" and type(val)(rec[key]) > val:
                    return True
                if operation == "<" and type(val)(rec[key]) < val:
                    return True
                return False
            except (ValueError, TypeError):
                return False
        return False

    if not res or not isinstance(res, list) or not len(res) > 0:
        return res
    if not isinstance(criteria, dict):
        return res
    for f, v in criteria.items():
        if not isinstance(v, dict):
            continue
        if f not in BTE_FILTERS:
            if "=" in v:
                res = list(
                    filter(
                        partial(filter_by_operation, key=f, val=v["="], operation="="),
                        res,
                    )
                )
                continue
            if ">" in v:
                res = list(
                    filter(
                        partial(filter_by_operation, key=f, val=v[">"], operation=">"),
                        res,
                    )
                )
            elif "<" in v:
                res = list(
                    filter(
                        partial(filter_by_operation, key=f, val=v["<"], operation="<"),
                        res,
                    )
                )
    return res



semmed = {
    "AnatomicalEntity": {
        "affected_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene",
            "CellularComponent"
        ],
        "affects": [
            "Disease",
            "BiologicalProcess"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "disrupted_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene"
        ],
        "manifestation_of": [
            "BiologicalProcess"
        ],
        "manifested_by": [
            "Disease"
        ],
        "negatively_regulated_by": [
            "ChemicalSubstance"
        ],
        "negatively_regulates": [
            "ChemicalSubstance",
            "Gene"
        ],
        "physically_interacts_with": [
            "ChemicalSubstance",
            "Gene"
        ],
        "positively_regulated_by": [
            "ChemicalSubstance",
            "Gene"
        ],
        "positively_regulates": [
            "ChemicalSubstance",
            "Gene"
        ],
        "preceded_by": [
            "Disease",
            "BiologicalProcess"
        ],
        "precedes": [
            "Disease",
            "BiologicalProcess"
        ],
        "predisposes": [
            "Disease"
        ],
        "prevents": [
            "Disease"
        ],
        "related_to": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ]
    },
    "BiologicalProcess": {
        "affected_by": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "affects": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "caused_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "disrupted_by": [
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "disrupts": [
            "AnatomicalEntity",
            "Cell",
            "BiologicalProcess",
            "Disease",
            "Gene",
            "CellularComponent"
        ],
        "negatively_regulated_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "negatively_regulates": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene"
        ],
        "physically_interacts_with": [
            "ChemicalSubstance",
            "Disease",
            "BiologicalProcess",
            "Gene"
        ],
        "positively_regulated_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene"
        ],
        "positively_regulates": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Gene"
        ],
        "prevents": [
            "Disease"
        ],
        "related_to": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "treats": [
            "Disease"
        ]
    },
    "Cell": {
        "affected_by": [
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "affects": [
            "Disease",
            "Cell"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "coexists_with": [
            "Cell",
            "ChemicalSubstance",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "derives_from": [
            "ChemicalSubstance"
        ],
        "disrupted_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "disrupts": [
            "BiologicalProcess"
        ],
        "has_part": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "located_in": [
            "CellularComponent",
            "AnatomicalEntity",
            "Cell"
        ],
        "location_of": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "manifestation_of": [
            "Disease",
            "BiologicalProcess"
        ],
        "manifested_by": [
            "Disease"
        ],
        "negatively_regulated_by": [
            "ChemicalSubstance"
        ],
        "negatively_regulates": [
            "ChemicalSubstance",
            "Gene"
        ],
        "part_of": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene",
            "CellularComponent"
        ],
        "physically_interacts_with": [
            "ChemicalSubstance",
            "Gene"
        ],
        "positively_regulated_by": [
            "ChemicalSubstance",
            "Gene"
        ],
        "positively_regulates": [
            "ChemicalSubstance",
            "Gene"
        ],
        "predisposes": [
            "Disease"
        ],
        "produced_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene",
            "CellularComponent"
        ],
        "produces": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "related_to": [
            "Disease",
            "Gene"
        ],
        "subclass_of": [
            "Disease",
            "BiologicalProcess"
        ],
        "treats": [
            "Disease"
        ]
    },
    "CellularComponent": {
        "affected_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "affects": [
            "AnatomicalEntity",
            "Cell",
            "BiologicalProcess",
            "Disease",
            "Gene",
            "CellularComponent"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "disrupted_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "disrupts": [
            "Disease",
            "BiologicalProcess",
            "Cell"
        ],
        "located_in": [
            "CellularComponent",
            "AnatomicalEntity",
            "Cell"
        ],
        "negatively_regulated_by": [
            "CellularComponent",
            "ChemicalSubstance",
            "Gene"
        ],
        "negatively_regulates": [
            "CellularComponent",
            "ChemicalSubstance",
            "Gene"
        ],
        "physically_interacts_with": [
            "CellularComponent",
            "ChemicalSubstance",
            "Disease",
            "Gene"
        ],
        "positively_regulated_by": [
            "CellularComponent",
            "ChemicalSubstance",
            "Gene"
        ],
        "positively_regulates": [
            "CellularComponent",
            "ChemicalSubstance",
            "Gene"
        ],
        "predisposes": [
            "Disease"
        ],
        "related_to": [
            "Disease",
            "AnatomicalEntity",
            "BiologicalProcess"
        ],
        "treats": [
            "Disease"
        ]
    },
    "ChemicalSubstance": {
        "affected_by": [
            "ChemicalSubstance",
            "Disease",
            "BiologicalProcess",
            "Gene"
        ],
        "affects": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "coexists_with": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "disrupted_by": [
            "ChemicalSubstance",
            "Gene"
        ],
        "disrupts": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "negatively_regulated_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "negatively_regulates": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "physically_interacts_with": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "positively_regulated_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "positively_regulates": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "prevents": [
            "Disease"
        ],
        "produced_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "produces": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "PhenotypicFeature",
            "Gene"
        ],
        "related_to": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "treats": [
            "Disease"
        ]
    },
    "Disease": {
        "affected_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "affects": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "caused_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "coexists_with": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "derives_from": [
            "ChemicalSubstance",
            "Gene"
        ],
        "derives_info": [
            "ChemicalSubstance"
        ],
        "disrupted_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "disrupts": [
            "AnatomicalEntity",
            "Cell",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "negatively_regulated_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "negatively_regulates": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene"
        ],
        "physically_interacts_with": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "positively_regulated_by": [
            "ChemicalSubstance",
            "Disease",
            "Gene"
        ],
        "positively_regulates": [
            "ChemicalSubstance",
            "Disease",
            "BiologicalProcess",
            "Gene"
        ],
        "prevented_by": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "prevents": [
            "Disease"
        ],
        "related_to": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "treated_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "treats": [
            "Disease"
        ]
    },
    "Gene": {
        "affected_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene",
            "CellularComponent"
        ],
        "affects": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "disrupted_by": [
            "ChemicalSubstance",
            "Disease",
            "BiologicalProcess",
            "Gene"
        ],
        "disrupts": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "negatively_regulated_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "negatively_regulates": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "physically_interacts_with": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "positively_regulated_by": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "positively_regulates": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene",
            "CellularComponent"
        ],
        "prevents": [
            "Disease"
        ],
        "related_to": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "treats": [
            "Disease"
        ]
    },
    "PhenotypicFeature": {
        "affected_by": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene"
        ],
        "affects": [
            "CellularComponent",
            "Disease",
            "BiologicalProcess",
            "Cell"
        ],
        "causes": [
            "Disease",
            "BiologicalProcess"
        ],
        "disrupted_by": [
            "ChemicalSubstance",
            "Disease",
            "Gene"
        ],
        "disrupts": [
            "CellularComponent",
            "Disease",
            "BiologicalProcess",
            "Cell"
        ],
        "has_part": [
            "ChemicalSubstance",
            "Disease",
            "CellularComponent",
            "Gene"
        ],
        "has_subclass": [
            "Disease",
            "PhenotypicFeature"
        ],
        "located_in": [
            "CellularComponent",
            "AnatomicalEntity",
            "Cell"
        ],
        "manifestation_of": [
            "Disease",
            "BiologicalProcess",
            "Gene"
        ],
        "manifested_by": [
            "Disease"
        ],
        "negatively_regulated_by": [
            "ChemicalSubstance",
            "Gene"
        ],
        "negatively_regulates": [
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene"
        ],
        "part_of": [
            "AnatomicalEntity",
            "Cell",
            "ChemicalSubstance",
            "Disease",
            "CellularComponent"
        ],
        "physically_interacts_with": [
            "ChemicalSubstance",
            "Disease",
            "Gene"
        ],
        "positively_regulated_by": [
            "ChemicalSubstance",
            "Gene"
        ],
        "positively_regulates": [
            "ChemicalSubstance",
            "Gene"
        ],
        "preceded_by": [
            "Disease",
            "BiologicalProcess"
        ],
        "precedes": [
            "BiologicalProcess"
        ],
        "predisposes": [
            "Disease"
        ],
        "prevents": [
            "Disease"
        ],
        "produced_by": [
            "Cell",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene",
            "CellularComponent"
        ],
        "produces": [
            "ChemicalSubstance",
            "Gene"
        ],
        "related_to": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "Gene"
        ],
        "subclass_of": [
            "AnatomicalEntity",
            "ChemicalSubstance",
            "BiologicalProcess",
            "Disease",
            "PhenotypicFeature",
            "Gene"
        ],
        "treats": [
            "Disease"
        ]
    }
}


from collections import Counter


class NodeDegreeFilter:
    def __init__(self, stepResult, criteria):
        self.stepResult = stepResult
        self.cnt = Counter()
        self.criteria = criteria

    def annotateNodeDegree(self):
        """
        Add node degree info to each edge.
        :param stepResult: a list of returned result from query.
        """
        if not self.stepResult:
            return
        for rec in self.stepResult:
            self.cnt[rec.get("$output")] += 1
        for rec in self.stepResult:
            rec["$nodeDegree"] = self.cnt[rec.get("$output")]
        return

    def filter(self):
        res = []
        if "sort" in self.criteria:
            limit = (
                int(self.criteria["sort"].get("limit"))
                if self.criteria["sort"].get("limit")
                else 0
            )
            if self.criteria["sort"].get("direction") != "descend":
                idList = [item[0] for item in self.cnt.most_common(limit)]
                for rec in self.stepResult:
                    if rec["$output"] in idList:
                        res.append(rec)
                return res
            sortedList = self.cnt.most_common(len(self.cnt))
            sortedList.reverse()
            idList = [item[0] for i, item in enumerate(sortedList) if i < limit]
            for rec in self.stepResult:
                if rec["$output"] in idList:
                    res.append(rec)
            return res



from .base_transformer import BaseTransformer


class OpenTargetTransformer(BaseTransformer):
    def wrap(self, res):
        if not res.get("data"):
            return res
        for _doc in res["data"]:
            if "drug" in _doc:
                if "CHEMBL" in _doc["drug"]["id"]:
                    _doc["drug"]["id"] = _doc["drug"]["id"].split("/")[-1]
        return res



# -*- coding: utf-8 -*-
"""
Resolving Biomedical Identifiers through BioThings APIs.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>
"""

from collections import defaultdict
from .config_new import ID_RESOLVING_APIS
from .apicall import BioThingsCaller


class IDResolver:
    """Resolving Biomedical Identifiers through BioThings APIs."""

    def __init__(self):
        """Load BTE registry and BTE API caller."""
        self.caller = BioThingsCaller()

    @staticmethod
    def fetch_id_mapping_file(semantic_type):
        """Fetch schema mapping file from the registry.

        :param: semantic_type: the id semantic type.
        """
        return ID_RESOLVING_APIS[semantic_type]["mapping"]

    @staticmethod
    def get_output_fields(mapping_file):
        fields = []
        for v in mapping_file.values():
            fields += v
        return ",".join(fields)

    @staticmethod
    def get_input_fields(mapping_file, _type):
        input_fields = mapping_file.get(_type)
        return ",".join(input_fields) if input_fields else None

    def resolve_ids(self, inputs, loop=None):
        """Main function to resolve identifiers.
        
        :param: inputs: A list of tuples, [(ids, prefix, semantic_type), ...]
        """
        # this parameter stores the final id conversion outputs
        self.results = {}
        self.api_call_inputs = []
        self.mapping_files = []
        self.types = []
        for _input in inputs:
            ids, _type, semantic_type = _input
            # convert id to list
            ids = self.preprocess_ids(ids, _type)
            # if _type == 'efo':
            # ids = [i.split(':')[-1] for i in ids]
            # if id can not be converted, the equivalent id is itself
            if semantic_type not in ID_RESOLVING_APIS:
                for _id in ids:
                    self.results[_type + ":" + _id] = {_type: [_id]}
            else:
                self.construct_api_calls(semantic_type, _type, ids)
        # make API calls asynchronously and gather all outputs
        self.responses, _ = self.caller.call_apis(self.api_call_inputs, loop=loop)
        self.parse_api_responses()
        return self.results

    def parse_api_responses(self):
        """Parse the API responses from APICall module in BTE."""
        for _res, _map, _type in zip(self.responses, self.mapping_files, self.types):
            # if API response is empty, continue
            if not _res:
                continue
            for single_res in _res["result"]:
                # if query of the item returns no hit
                res_id = _type + ":" + single_res["query"]
                if "notfound" in single_res:
                    self.results[res_id] = {_type: [single_res["query"]]}
                    continue
                if res_id not in self.results:
                    self.results[res_id] = defaultdict(set)
                for k, v in _map.items():
                    for _v in v:
                        if _v in single_res:
                            val = single_res[_v]
                            if not isinstance(val, list):
                                self.results[res_id][k].add(str(val))
                            else:
                                self.results[res_id][k].update(set(val))
            for res_id, resolved_ids in self.results.items():
                for m in resolved_ids.keys():
                    if m == "name":
                        self.results[res_id][m] = sorted(
                            {(str(item)).upper() for item in resolved_ids[m]}
                        )
                    else:
                        self.results[res_id][m] = list(resolved_ids[m])

    def construct_api_calls(self, semantic_type, id_type, ids):
        """Construct API calls for BTE API call module.

        :param: api: name of API
        :param: _type: the prefix type of ids
        :param: ids: the list of identifiers to resolve
        """
        mapping_file = self.fetch_id_mapping_file(semantic_type)
        api = ID_RESOLVING_APIS[semantic_type]["api_name"]
        if self.get_input_fields(mapping_file, id_type):
            for i in range(0, len(ids), 1000):
                self.api_call_inputs.append(
                    {
                        "api": api,
                        "operation": {
                            "server": ID_RESOLVING_APIS[semantic_type]["url"],
                            "path": "/query",
                            "method": "post",
                            "parameters": {
                                "fields": self.get_output_fields(mapping_file),
                                "dotfield": "true",
                                "species": "human",
                            },
                            "requestBody": {
                                "body": {
                                    "q": "{inputs[0]}",
                                    "scopes": self.get_input_fields(
                                        mapping_file, id_type
                                    ),
                                },
                                "header": "application/x-www-form-urlencoded",
                            },
                        },
                        "value": ",".join(ids[i : i + 1000]),
                        "internal_query_id": 1,
                    }
                )
                self.types.append(id_type)
                self.mapping_files.append(mapping_file)
        else:
            for _id in ids:
                self.results[id_type + ":" + _id] = {id_type: [_id]}

    def preprocess_ids(self, ids, _type):
        """Preprocess ids to become a list of strings.

        :param: ids: list of input ids
        :param: _type: the prefix type
        """
        # convert id to list
        if isinstance(ids, str):
            ids = [ids]
        # make sure all ids in id list is str
        for _id in ids:
            if " " in str(_id):
                self.results[_type + ":" + str(_id)] = {_type: [str(_id)]}
        ids = [str(i) for i in ids if " " not in str(i)]
        return ids



from .transformers.automat_transformer import AutomatTransformer
from .transformers.base_transformer import BaseTransformer
from .transformers.biolink_transformer import BioLinkTransformer
from .transformers.biothings_transformer import BioThingsTransformer
from .transformers.cord_transformer import CordTransformer
from .transformers.ctd_transformer import CTDTransformer
from .transformers.opentarget_transformer import OpenTargetTransformer
from .transformers.semmed_transformer import SemmedTransformer
from .transformers.mut_freq_transformer import MutFreqTransformer
from .transformers.drug_response_transformer import DrugResponseTransformer
from .transformers.reasoner_transformer import ReasonerTransformer
from .transformers.cohd_transformer import COHDTransformer


class Transformer:
    def __init__(self, data):
        self.data = data
        self.route()

    def route(self):
        api = self.data["edge"]["association"]["api_name"]
        tags = self.data["edge"]["query_operation"]["tags"]
        if api.startswith("CORD"):
            self.tf = CordTransformer(self.data)
        elif api.startswith("SEMMED"):
            self.tf = SemmedTransformer(self.data)
        elif api == "TCGA Mutation Frequency KP API":
            self.tf = MutFreqTransformer(self.data)
        elif api == "Drug Response KP API":
            self.tf = DrugResponseTransformer(self.data)
        elif api == "BioLink API":
            self.tf = BioLinkTransformer(self.data)
        elif "automat" in tags:
            self.tf = AutomatTransformer(self.data)
        elif "biothings" in tags:
            self.tf = BioThingsTransformer(self.data)
        elif "reasoner" in tags:
            self.tf = ReasonerTransformer(self.data)
        elif "ctd" in tags:
            self.tf = CTDTransformer(self.data)
        elif "opentarget" in tags:
            self.tf = OpenTargetTransformer(self.data)
        elif api == "COHD API":
            self.tf = COHDTransformer(self.data)
        else:
            self.tf = BaseTransformer(self.data)

    def transform(self):
        return self.tf.transform()



# -*- coding: utf-8 -*-
"""
API-specific JSON response restructure.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""

from .biolink import restructure_biolink_response
from .reasoner import restructure_reasoner_response
from .stanford import restructure_stanford_response
from .ctd import restructure_ctd_response
from .opentarget import restructure_opentarget_response
from .semmed import restructure_semmed_response
from .cord import restructure_cord_response
from .scibite import restructure_scibite_response

class APIPreprocess():

    """
    Restructure the output of specific APIs.
    
    This is to make sure the JSON output from these APIs could be uniformly consumed by the jsontransform module
    """

    def __init__(self, json_doc, api_type, api_name=None, output_types=None):
        """
        Load json doc and api info.

        :param: json_doc: the json_doc to be preprocessed
        :param: api_type: the type of api, e.g. biothings
        :param: api_name: optional, the name of api
        :param: output_types: specific for semmed api

        """
        self.api_type = api_type
        self.api_name = api_name
        self.json_doc = json_doc
        self.output_types = output_types

    def restructure(self):
        """Restructue API response."""
        # if input is empty, do not restructure
        if not self.json_doc:
            return self.json_doc
        # if input is a list, turn it into a dictionary
        if isinstance(self.json_doc, list) and self.api_type != "biothings":
            self.json_doc = {"data": self.json_doc}
        if self.api_type == 'biolink':
            return restructure_biolink_response(self.json_doc)
        if self.api_type == 'reasoner':
            return restructure_reasoner_response(self.json_doc)
        if self.api_type == 'stanford':
            return restructure_stanford_response(self.json_doc)
        if self.api_type == 'ctd':
            return restructure_ctd_response(self.json_doc)
        if self.api_type == 'opentarget':
            return restructure_opentarget_response(self.json_doc)
        if self.api_name[:4] == 'semm':
            return restructure_semmed_response(self.json_doc, self.output_types)
        if self.api_name[:4] == 'cord':
            return restructure_cord_response(self.json_doc, self.output_types)
        if self.api_name in ['scibite', 'scigraph', 'pharos', 'hmdb', 'hetio', 'chembio']:
            return restructure_scibite_response(self.json_doc)
        return self.json_doc




# -*- coding: utf-8 -*-

"""Accept user query and return results as a graph.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>

"""
from collections import defaultdict
import networkx as nx
import copy

from .api_call_dispatcher import Dispatcher
from .id_resolver import IDResolver
from .registry import Registry
from .utils.networkx import (
    load_res_to_networkx,
    add_equivalent_ids_to_nodes,
    merge_two_networkx_graphs,
)
from .utils.common import dict2listoftuples, listoftuples2dict
from .utils.common import (
    get_name_from_equivalent_ids,
    get_primary_id_from_equivalent_ids,
)
from .metadata import Metadata
from .export.reasoner import ReasonerConverter
from .export.graphml import GraphmlConverter
from .export.pandas import networkx2pandas
from .export.graphviz import networkx2graphvis

# TODO: remove ID_RANK
ID_RANK = {
    "Gene": "SYMBOL",
    "ChemicalSubstance": "name",
    "DiseaseOrPhenotypicFeature": "name",
}


class SingleEdgeQueryDispatcher:
    """Query from one bio-entity to other type(s) of bio-entities.

    params
    ------
    input_cls: str or list, the semantic type(s) of the input, e.g. Gene
    input_id: str, the identifier type of the input, e.g. entrez
    values: str or list, the actual value of the input, e.g. CDK7
    output_cls: str or list or None, the semantic type(s) of the output, e.g. Gene. If None, search for all semantic type(s) which connect from the input
    output_id: str, optional, the identifier type in which output ids will be, e.g. chembl
    pred: str or list or None, the relationship between input and output. If None, search for all relationships between input and output
    input_obj: optional, a representation of the input from Hint Module
    prev_graph: the graph object coming from last query
    registry: optional, the Registry object in BioThings Explorer
    """

    def __init__(
        self,
        input_cls=None,
        input_id=None,
        values=None,
        output_cls=None,
        output_id=None,
        pred=None,
        equivalent_ids=None,
        input_obj=None,
        prev_graph={},
        query_id=1,
        reverse=False,
        registry=None,
        loop=None,
    ):
        self.loop = loop
        # load bte registry
        if not registry:
            self.registry = Registry()
        else:
            self.registry = registry
        self.prev_graph = prev_graph
        self.query_id = query_id
        self.metadata = Metadata(reg=self.registry)
        # load id conversion module
        self.idr = IDResolver()
        semantic_types = self.metadata.list_all_semantic_types()
        id_types = self.metadata.list_all_id_types()
        self.input_cls = input_cls
        self.input_id = input_id
        if output_cls in (["BiologicalEntity"], "BiologicalEntity"):
            self.output_cls = None
        else:
            self.output_cls = output_cls
        if isinstance(self.output_cls, str):
            self.output_cls = [self.output_cls]
        self.output_id = output_id
        # if output id is not specified, use the default id for this type
        if not output_id:
            if (
                self.output_cls
                and not isinstance(self.output_cls, list)
                and self.output_cls in ID_RANK
            ):
                self.output_id = [ID_RANK.get(self.output_cls)]
            else:
                self.output_id = ["SYMBOL", "name"]
        if not isinstance(self.output_id, list):
            self.output_id = [self.output_id]
        self.pred = pred
        self.values = values
        self.equivalent_ids = equivalent_ids
        self.input_identifier = None
        self.reverse = reverse
        if input_obj:
            self.input_cls = input_obj.get("primary").get("cls")
            self.input_id = input_obj.get("primary").get("identifier")
            self.values = input_obj.get("primary").get("value")
            if "SYMBOL" in input_obj:
                self.input_label = input_obj.get("SYMBOL")
                self.input_identifier = "SYMBOL"
            elif "name" in input_obj:
                self.input_label = input_obj.get("name")
                self.input_identifier = "name"
            else:
                self.input_label = (
                    input_obj.get("primary").get("identifier")
                    + ":"
                    + input_obj.get("primary").get("value")
                )
                self.input_identifier = input_obj.get("primary").get("identifier")
        else:
            self.input_label = None
        # check if input_cls is valid
        if self.input_cls not in semantic_types:
            raise Exception(
                "The input_cls is not valid. Valid input classes are {}".format(
                    semantic_types
                )
            )
        if not self.equivalent_ids:
            # find equivalent ids for the input value
            equivalent_ids = self.idr.resolve_ids(
                [(self.values, self.input_id, self.input_cls)], loop=self.loop
            )
            if not self.input_label:
                self.input_label = self.input_id + ":" + values
            self.equivalent_ids = equivalent_ids
        self.dp = Dispatcher(registry=self.registry)
        self.G = nx.MultiDiGraph()
        self.log = []
        self.summary = []

    def to_reasoner_std(self):
        """Convert the output to reasoner api standard.
        """
        rc = ReasonerConverter()
        rc.load_bte_query_path(
            start={
                "type": self.input_cls,
                "primary": {"identifier": self.input_id, "value": self.values},
            },
            intermediate=None,
            end={"type": self.output_cls[0]},
        )
        rc.load_bte_output(self.G)
        return rc.generate_reasoner_response()

    @staticmethod
    def group_edges_by_input_id(edges):
        """Group edges based on the input identifier.

        :param: edges: list of dicts, edges filtered based on input/output/pred.
        """
        if not edges:
            return edges
        grouped_edges = defaultdict(list)
        for _edge in edges:
            # need to convert to tuple to make it immutable
            grouped_edges[_edge["operation"]["input_id"]].append(
                dict2listoftuples(_edge)
            )
        return grouped_edges

    def merge_equivalent_nodes(self):
        """Merge equivalent nodes together.

        nodes will be merged based on their equivalent ids
        edges will be added to the merged node
        """
        if self.G:
            nodes_to_remove = set()
            nodes_to_add = []
            edges_to_add = []
            identifiers = []
            # loop through all edges
            # n1 is the subject node, n2 is the object node
            for n1, n2, data in self.G.edges(data=True):
                # get the id type of the object node
                identifier = self.G.nodes[n2]["identifier"]
                # check if the identifier matches the id type specified by the
                # user or the default id type
                if self.output_id and identifier not in self.output_id:
                    # if equivalent_ids doesn't exist then print error message
                    if "equivalent_ids" not in self.G.nodes[n2]:
                        print("ID resolution steps failed:")
                    equivalent_ids = self.G.nodes[n2]["equivalent_ids"]
                    # find the corresponding id from the equivalent id dict
                    new_vals = None
                    for _id in self.output_id:
                        if equivalent_ids.get(_id):
                            if _id == "name":
                                new_vals = [equivalent_ids.get(_id)[0]]
                            else:
                                new_vals = equivalent_ids.get(_id)
                            break
                    if new_vals:
                        # get n2's node info
                        node_info = self.G.nodes[n2]
                        # change n2's identifier
                        for _val in new_vals:
                            # add new edge
                            edges_to_add.append((n1, _val, data))
                            # set the original node to be removed
                            nodes_to_remove.add(n2)
                            # if this node is not in graph, add this node
                            nodes_to_add.append((_val, node_info))
                            identifiers.append(_id)
            # remove duplicate nodes
            for n in nodes_to_remove:
                self.G.remove_node(n)
            # update identifier type
            for k, n in zip(identifiers, nodes_to_add):
                n[1]["identifier"] = k
            # add new nodes and edges
            self.G.add_nodes_from(nodes_to_add)
            self.G.add_edges_from(edges_to_add)

    def construct_internal_graph(self, reverse=False):
        graph = defaultdict(list)
        if reverse:
            g = self.G.reverse(copy=True)
        else:
            g = self.G
        for sbj, obj, info in g.edges(data=True):
            sbj_name = get_name_from_equivalent_ids(
                g.nodes[sbj]["equivalent_ids"], None
            )
            sbj_id = get_primary_id_from_equivalent_ids(
                g.nodes[sbj]["equivalent_ids"], g.nodes[sbj]["type"]
            )
            obj_name = get_name_from_equivalent_ids(
                g.nodes[obj]["equivalent_ids"], None
            )
            obj_id = get_primary_id_from_equivalent_ids(
                g.nodes[obj]["equivalent_ids"], g.nodes[obj]["type"]
            )
            sbj_label = (
                str(self.query_id - 1)
                + "-"
                + g.nodes[sbj]["identifier"]
                + ":"
                + str(sbj).upper()
            )
            obj_label = (
                str(self.query_id)
                + "-"
                + g.nodes[obj]["identifier"]
                + ":"
                + str(obj).upper()
            )
            tmp = []
            if sbj_label in self.prev_graph:
                prev_graph = copy.deepcopy(self.prev_graph[sbj_label])
                for prec_rec in prev_graph:
                    prec_rec.append(
                        {
                            "input": sbj_label,
                            "output": obj_label,
                            "info": info,
                            "input_name": sbj_name,
                            "input_id": sbj_id,
                            "output_name": obj_name,
                            "output_id": obj_id,
                        }
                    )
                    tmp.append(prec_rec)
            else:
                if not reverse:
                    tmp = [
                        [
                            {
                                "input": sbj_label,
                                "output": obj_label,
                                "info": info,
                                "input_name": sbj_name,
                                "input_id": sbj_id,
                                "output_name": obj_name,
                                "output_id": obj_id,
                            }
                        ]
                    ]
            graph[obj_label] += tmp
        self.current_graph = graph

    def query(self, verbose=False):
        """Query APIs and organize outputs into networkx graph."""
        self.current_graph = {}
        output_cls_name = " AND ".join(self.output_cls) if self.output_cls else "None"
        if verbose:
            print("==== Step #1: Query path planning ====")
            print(
                "\nBecause {} is of type '{}', BTE will query our meta-KG for APIs that can take '{}' as input and '{}' as output".format(
                    self.input_label, self.input_cls, self.input_cls, output_cls_name
                )
            )
        self.log.append("==== Step #1: Query path planning ====")
        self.log.append(
            "\nBecause {} is of type '{}', BTE will query our meta-KG for APIs that can take '{}' as input and '{}' as output".format(
                self.input_label, self.input_cls, self.input_cls, output_cls_name
            )
        )
        # filter edges based on subject, object, predicate
        edges = self.registry.filter_edges(self.input_cls, self.output_cls, self.pred)
        if not edges:
            # print("No edges found for the <input, pred, output> you specified")
            if verbose:
                print(
                    "We are sorry! We couln't find any APIs which can do the type of query for you!"
                )
            self.log.append(
                "We are sorry! We couln't find any APIs which can do the type of query for you!"
            )
            return

        grouped_edges = self.group_edges_by_input_id(edges)
        # t1 = time.time()
        # print("equivalent_ids", self.equivalent_ids)
        # t2 = time.time()
        # print('time to find equivalent ids {}'.format(t2-t1))
        input_edges = []
        mapping_keys = []
        output_id_types = []
        id_mapping = {}
        # populate edge with corresponding input value
        # k is the source id, v is the equivalent ids
        for k, v in self.equivalent_ids.items():
            # p is the input id, q is the edges from the source id
            for p, q in grouped_edges.items():
                # check if input id is in equivalent ids
                if p in v and v[p]:
                    for _edge in q:
                        m = _edge
                        m = listoftuples2dict(m)
                        m["value"] = v[p]
                        mapping_keys.append(m["label"])
                        input_edges.append(m)
                        output_id_types.append(m["operation"]["output_id"])
                    input_identifier = (
                        self.input_identifier
                        if self.input_identifier
                        else k.split(":", 1)[0]
                    )
                    self.G.add_node(
                        get_name_from_equivalent_ids(v, self.input_label),
                        type=self.input_cls,
                        identifier=input_identifier,
                        level=1,
                        equivalent_ids=self.equivalent_ids[k],
                    )
                    for _id in v[p]:
                        id_mapping[_id] = get_name_from_equivalent_ids(
                            v, self.input_label
                        )
        if not input_edges:
            if verbose:
                print(
                    "We are sorry! We couln't find any APIs which can do the type of query for you!"
                )
            self.log.append(
                "We are sorry! We couln't find any APIs which can do the type of query for you!"
            )
            return
        source_nodes_cnt = len(self.G)
        # make API calls and restructure API outputs
        (_res, log) = self.dp.dispatch(input_edges, verbose=verbose, loop=self.loop)
        self.log += log
        # load API outputs into the MultiDiGraph
        self.G = load_res_to_networkx(
            _res, self.G, mapping_keys, id_mapping, output_id_types
        )
        # annotate nodes with its equivalent ids
        self.G, out_equ_ids = add_equivalent_ids_to_nodes(self.G, self.idr)
        self.equivalent_ids.update(out_equ_ids)
        # merge equivalent nodes
        self.merge_equivalent_nodes()
        self.construct_internal_graph(reverse=self.reverse)
        if verbose:
            print(
                "\nAfter id-to-object translation, BTE retrieved {} unique objects.".format(
                    len(self.G) - source_nodes_cnt
                )
            )
        self.log.append(
            "\nAfter id-to-object translation, BTE retrieved {} unique objects.".format(
                len(self.G) - source_nodes_cnt
            )
        )

    def to_json(self):
        """convert the graph into JSON through networkx"""
        # first check if the graph is empty
        if self.G.number_of_nodes() > 0:
            res = nx.json_graph.node_link_data(self.G)
            return res
        else:
            return {}

    @property
    def output_ids(self):
        """Retrieve output ids along with their equivalent ids."""
        result = {}
        if self.G.number_of_nodes() > 0:
            for x, y in self.G.nodes(data=True):
                if y["level"] and y["level"] == 2:
                    identifier = y["identifier"]
                    semantic_type = y["type"]
                    if semantic_type not in result:
                        result[semantic_type] = {}
                    curie = identifier + ":" + x
                    result[semantic_type][curie] = y["equivalent_ids"]
        return result

    def show_all_nodes(self):
        """show all nodes in the graph"""
        return list(self.G.nodes())

    def show_all_edges(self):
        """show all edges in the graph"""
        return list(self.G.edges())

    def display_node_info(self, node):
        """show detailed node information

        Params
        ------
        node: str, node id
        """
        if node not in self.G:
            raise Exception("{} is not in the graph".format(node))
        return self.G.nodes[node]

    def display_edge_info(self, start_node, end_node):
        """display detailed edge info between start node and end node

        Params
        ------
        start_node: str, start node id
        end_node: str, end node id
        """
        if start_node not in self.G:
            raise Exception("{} is not in the graph".format(start_node))
        if end_node not in self.G:
            raise Exception("{} is not in the graph".format(end_node))
        if not self.G.has_edge(start_node, end_node):
            raise Exception(
                "No edge exists between {} and {}".format(start_node, end_node)
            )
        return dict(self.G[start_node][end_node])

    def display_table_view(self):
        return networkx2pandas(self.G, self.input_cls)


class Explain:
    """Find intermediate node(s) that connect input and output
    
    There might not be direct evidence showing connection between A and B. \
        The "FindConnection" class aims at finding intermediate node(s) which both A and B are connected to.

    Parameters:
        input_obj: dict
            the input object returned from Hint, required
        output_obj: dict
            the output object returned from Hint, required
        intermediate_nodes: dict
            the semantic type(s) of the intermediate node
            could be None, which represents any semantic type, or a list of semantic types
        
    Examples
        >>> fc = FindConnection(input_obj=asthma, output_obj=imatinib, intermediate_nodes=[None])
        >>> fc.connect()

        >>> fc = FindConnection(input_obj=asthma, output_obj=imatinib, intermediate_nodes=['Gene'])
        >>> fc.connect()

        >>> fc = FindConnection(input_obj=asthma, output_obj=imatinib, intermediate_nodes=['Gene', 'ChemicalSubstance'])
        >>> fc.connect()

    """

    def __init__(self, input_obj, output_obj, intermediate_nodes=None, registry=None):
        if not registry:
            self.registry = Registry()
        else:
            self.registry = registry
        # convert intermediate_cls into a list if not
        if not intermediate_nodes:
            self.intermediate_cls = output_obj["type"]
            self.intermediate_cls_cnt = 0
        elif isinstance(intermediate_nodes, list):
            self.intermediate_cls = intermediate_nodes
            self.intermediate_cls_cnt = len(self.intermediate_cls)
        else:
            self.intermediate_cls = [intermediate_nodes]
            self.intermediate_cls_cnt = len(self.intermediate_cls)
        self.input_obj = input_obj
        # self.starts represents the display label of the input
        if "SYMBOL" in input_obj:
            self.starts = input_obj["SYMBOL"]
        elif "name" in input_obj:
            self.starts = input_obj["name"]
        else:
            self.starts = self.input_obj.get("primary").get(
                "identifier"
            ) + self.input_obj.get("primary").get("value")
        self.output_obj = output_obj
        # self.ends represents the display label of the output
        if "SYMBOL" in output_obj:
            self.ends = output_obj["SYMBOL"]
        elif "name" in output_obj:
            self.ends = output_obj["name"]
        else:
            self.ends = self.output_obj.get("primary").get(
                "identifier"
            ) + self.output_obj.get("primary").get("value")
        self.G = nx.MultiDiGraph()
        self.seqd = {}
        self.current_graph = {}
        self.log = []

    def connect(self, verbose=False):
        if verbose:
            print("==========")
            print("========== QUERY PARAMETER SUMMARY ==========")
            print("==========\n")
            print(
                "BTE will find paths that join '{}' and '{}'. Paths will have {} intermediate node.\n".format(
                    self.starts, self.ends, self.intermediate_cls_cnt
                )
            )
            if self.intermediate_cls_cnt > 0:
                for i, item in enumerate(self.intermediate_cls):
                    print(
                        "Intermediate node #{} will have these type constraints: {}\n\n".format(
                            i + 1, ",".join([str(j) for j in self.intermediate_cls])
                        )
                    )
            print("==========")
        self.log.append("==========")
        self.log.append("========== QUERY PARAMETER SUMMARY ==========")
        self.log.append("==========\n")
        self.log.append(
            "BTE will find paths that join '{}' and '{}'. Paths will have {} intermediate node.\n".format(
                self.starts, self.ends, self.intermediate_cls_cnt
            )
        )
        if self.intermediate_cls_cnt > 0:
            for i, item in enumerate(self.intermediate_cls):
                self.log.append(
                    "Intermediate node #{} will have these type constraints: {}\n\n".format(
                        i + 1, ",".join([str(j) for j in self.intermediate_cls])
                    )
                )
        self.log.append("==========")
        if self.intermediate_cls_cnt == 0:
            output_cls = self.intermediate_cls
        elif self.intermediate_cls == ["BiologicalEntity"]:
            output_cls = "Biological Entities"
        else:
            output_cls = " AND ".join(self.intermediate_cls) + " entities"
        if verbose:
            print(
                "========== QUERY #1 -- fetch all {} linked to '{}' ==========".format(
                    output_cls, self.starts
                )
            )
            print("==========\n")
        self.log.append(
            "========== QUERY #1 -- fetch all {} linked to '{}' ==========".format(
                output_cls, self.starts
            )
        )
        self.log.append("==========\n")
        self.seqd[1] = SingleEdgeQueryDispatcher(
            input_obj=self.input_obj,
            output_cls=self.intermediate_cls,
            query_id=1,
            registry=self.registry,
        )
        self.seqd[1].query(verbose=verbose)
        self.log += self.seqd[1].log
        if len(self.seqd[1].G) < 2:
            print(
                "The first query doesn't return any result. So BTE does not find any connection between your input and output"
            )
            return
        self.G = copy.deepcopy(self.seqd[1].G)
        if self.intermediate_cls_cnt == 0:
            if self.ends in self.G:
                found = ""
            else:
                found = "did not "
            if verbose:
                print("\n==========")
                print("========== Final assembly of results ==========")
                print("==========\n\n")
                print(
                    "BTE {}found direct connection between '{}' and '{}'".format(
                        found, self.starts, self.ends
                    )
                )
            self.log.append("\n==========")
            self.log.append("========== Final assembly of results ==========")
            self.log.append("==========\n\n")
            self.log.append(
                "BTE {}found direct connection between '{}' and '{}'".format(
                    found, self.starts, self.ends
                )
            )
        else:
            if verbose:
                print("\n\n==========")
                print(
                    "========== QUERY #2 -- fetch all {} linked to '{}' ==========".format(
                        output_cls, self.ends
                    )
                )
                print("==========\n")
            self.log.append("\n\n==========")
            self.log.append(
                "========== QUERY #2 -- fetch all {} linked to '{}' ==========".format(
                    output_cls, self.ends
                )
            )
            self.log.append("==========\n")
            self.seqd[2] = SingleEdgeQueryDispatcher(
                input_obj=self.output_obj,
                output_cls=self.intermediate_cls,
                prev_graph=self.seqd[1].current_graph,
                query_id=2,
                reverse=True,
                registry=self.registry,
            )
            self.seqd[2].query(verbose=verbose)
            if len(self.seqd[2].G) < 2:
                print(
                    "The second query doesn't return any result. So BTE does not find any connection between your input and output"
                )
                return
            self.log += self.seqd[2].log
            self.seqd[2].G = self.seqd[2].G.reverse()
            self.G = merge_two_networkx_graphs(self.G, self.seqd[2].G)
            self.sub_G = self.sub_graph()
            self.current_graph = self.seqd[2].current_graph
            if len(self.sub_G) <= 2:
                intermediate_cnt = 0
            else:
                intermediate_cnt = len(self.sub_G) - 2
            if verbose:
                print("\n==========")
                print("========== Final assembly of results ==========")
                print("==========\n\n")
                print(
                    "BTE found {} unique intermediate nodes connecting '{}' and '{}'".format(
                        intermediate_cnt, self.starts, self.ends
                    )
                )
            self.log.append("\n==========")
            self.log.append("========== Final assembly of results ==========")
            self.log.append("==========\n\n")
            self.log.append(
                "BTE found {} unique intermediate nodes connecting '{}' and '{}'".format(
                    intermediate_cnt, self.starts, self.ends
                )
            )

    def summary(self, attr):
        """"""

    def show_path(self, remove_duplicate=True):
        if remove_duplicate:
            paths = set()
            for path in nx.all_simple_paths(
                self.G, source=self.starts, target=self.ends
            ):

                path = "||".join(path)
                paths.add(path)
            new_paths = []
            for _path in paths:
                new_paths.append(_path.split("||"))
            return new_paths
        else:
            paths = []
            for path in nx.all_simple_paths(
                self.G, source=self.starts, target=self.ends
            ):
                paths.append(path)
        return paths

    def sub_graph(self):
        paths = self.show_path()
        nodes = set()
        for _path in paths:
            if _path:
                nodes = nodes | set(_path)
        return self.G.subgraph(nodes)

    def visualize(self):
        """Convert the graph to visjs JSON format."""
        return networkx2graphvis(self.sub_G)

    def to_json(self):
        """Convert the graph into JSON through networkx."""
        if self.sub_G.number_of_nodes() > 0:
            res = nx.json_graph.node_link_data(self.sub_G)
            return res
        return {}

    def show_all_nodes(self):
        """Show all nodes in the graph."""
        return list(self.sub_G.nodes())

    def show_all_edges(self):
        """show all edges in the graph"""
        return list(self.sub_G.edges())

    def display_node_info(self, node):
        """Show detailed node information.

        :param: node: node id
        """
        if node not in self.G:
            raise Exception("{} is not in the graph".format(node))
        return self.G.nodes[node]

    def display_edge_info(self, start_node, end_node):
        """Display detailed edge info between start node and end node.

        :param: start_node: str, start node id
        :param: end_node: str, end node id
        """
        if start_node not in self.G:
            raise Exception("{} is not in the graph".format(start_node))
        if end_node not in self.G:
            raise Exception("{} is not in the graph".format(end_node))
        if not self.G.has_edge(start_node, end_node):
            raise Exception(
                "No edge exists between {} and {}".format(start_node, end_node)
            )
        return dict(self.G[start_node][end_node])

    def display_table_view(self):
        """Display the query results as a pandas table.
        
        Examples
        --------
        >>> df = fc.display_table_view()
        >>> df

        """
        return networkx2pandas(self.current_graph, self.input_obj["type"])


class Predict:
    """Find relationships between one specific entity and a class of entity types.
    
    params
    ------
    input_obj: the input object returned from Hint, required
    output_obj: the class of entities as output, required
        could be None, str, or a list of entity classes
    intermediate_nodes: the semantic type(s) of the intermediate node
        could be None, which represents any semantic type, or a list of semantic types
    
    """

    def __init__(self, input_obj, output_obj, intermediate_nodes, registry=None):
        """Initialize.
        params
        ------
        input_obj: the input object returned from Hint, required
        output_obj: the class of entities as output, required
            could be None, str, or a list of entity classes
        intermediate_nodes: the semantic type(s) of the intermediate node
            could be None, which represents any semantic type, or a list of semantic types
        
        """
        if not intermediate_nodes:
            intermediate_nodes = []
        elif not isinstance(intermediate_nodes, list):
            intermediate_nodes = [intermediate_nodes]
        self.intermediate_nodes = copy.deepcopy(intermediate_nodes)
        self.paths = copy.deepcopy(intermediate_nodes)
        # append output_obj to the path
        self.paths.append(output_obj)
        self.input_obj = input_obj
        if "SYMBOL" in input_obj:
            self.starts = input_obj["SYMBOL"]
        elif "name" in input_obj:
            self.starts = input_obj["name"]
        else:
            self.starts = self.input_obj.get("primary").get(
                "identifier"
            ) + self.input_obj.get("primary").get("value")
        self.ends = output_obj
        if not registry:
            self.registry = Registry()
        else:
            self.registry = registry
        self.G = nx.MultiDiGraph()
        self.seqd = {}
        # aggregate output_ids from different queries
        self.output_ids = {}
        self.current_graph = {}
        self.prev_graph = {}
        self.log = []

    def merge_output_ids(self, query_id, output_ids):
        level = query_id.split(".")[0]
        for _type, _ids in output_ids.items():
            if _type in self.output_ids[level]:
                self.output_ids[level][_type].update(_ids)
            else:
                self.output_ids[level][_type] = _ids

    def connect(self, verbose=False):
        """Make the query.
        
        params
        ------
        verbose: boolean, if True, print the progress and the summary
        """
        if verbose:
            print("==========")
            print("========== QUERY PARAMETER SUMMARY ==========")
            print("==========\n")
            print(
                "BTE will find paths that join '{}' and '{}'. Paths will have {} intermediate node.\n".format(
                    self.starts, self.ends, len(self.intermediate_nodes)
                )
            )
            for i, item in enumerate(self.intermediate_nodes):
                print(
                    "Intermediate node #{} will have these type constraints: {}\n".format(
                        i + 1, item
                    )
                )
        self.log.append("==========")
        self.log.append("========== QUERY PARAMETER SUMMARY ==========")
        self.log.append("==========\n")
        self.log.append(
            "BTE will find paths that join '{}' and '{}'. Paths will have {} intermediate node.\n".format(
                self.starts, self.ends, len(self.intermediate_nodes)
            )
        )
        for i, item in enumerate(self.intermediate_nodes):
            self.log.append(
                "Intermediate node #{} will have these type constraints: {}\n".format(
                    i + 1, item
                )
            )
        for i, output_cls in enumerate(self.paths):
            if i == 0:
                # if it's the first element in the path
                # the input should be the user provided input
                equivalent_ids = None
                input_obj = self.input_obj
                input_cls = input_obj["type"]
            else:
                # if it's not the first element in the path
                # the input should be the results from the previous query
                input_cls = self.paths[i - 1]
                equivalent_ids = copy.deepcopy(self.output_ids[str(i)])
                input_obj = None
            if (i > 0 and equivalent_ids != {}) or i == 0:
                if input_obj:
                    _input = copy.copy(self.starts)
                if output_cls == "BiologicalEntity":
                    _output = "Biological Entities"
                else:
                    if not isinstance(output_cls, list) or not isinstance(
                        output_cls, tuple
                    ):
                        output_cls = [output_cls]
                    _output = " AND ".join(output_cls) + " entities"
                self.output_ids[str(i + 1)] = {}
                if equivalent_ids:
                    for j, input_cls in enumerate(equivalent_ids.keys()):
                        query_id = str(i + 1) + "." + str(j + 1)
                        if verbose:
                            print(
                                "\n\n========== QUERY #{} -- fetch all {} linked to {} entites ==========".format(
                                    query_id, _output, input_cls
                                )
                            )
                            print("==========\n")
                        self.log.append(
                            "\n\n========== QUERY #{} -- fetch all {} linked to {} entites ==========".format(
                                query_id, _output, input_cls
                            )
                        )
                        self.log.append("==========\n")
                        self.seqd[query_id] = SingleEdgeQueryDispatcher(
                            input_obj=input_obj,
                            equivalent_ids=equivalent_ids[input_cls],
                            input_cls=input_cls,
                            output_cls=output_cls,
                            query_id=i + 1,
                            prev_graph=self.prev_graph,
                            pred=None,
                        )
                        self.seqd[query_id].query(verbose=verbose)
                        if len(self.seqd[query_id].G) < 2:
                            break
                        self.log += self.seqd[query_id].log
                        # print(seqd.G.nodes())
                        self.G = merge_two_networkx_graphs(
                            self.G, self.seqd[query_id].G
                        )
                        self.merge_output_ids(query_id, self.seqd[query_id].output_ids)
                        self.prev_graph.update(self.seqd[query_id].current_graph)
                        if i + 1 == len(self.paths):
                            self.current_graph.update(self.seqd[query_id].current_graph)
                    else:
                        continue
                    break
                else:
                    if verbose:
                        print(
                            "\n\n========== QUERY #{} -- fetch all {} linked to {} ==========".format(
                                i + 1, _output, _input
                            )
                        )
                        print("==========\n")
                    self.log.append(
                        "\n\n========== QUERY #{} -- fetch all {} linked to {} ==========".format(
                            i + 1, _output, _input
                        )
                    )
                    self.log.append("==========\n")
                    self.seqd[i + 1] = SingleEdgeQueryDispatcher(
                        input_obj=input_obj,
                        input_cls=input_cls,
                        output_cls=output_cls,
                        query_id=i + 1,
                        prev_graph=self.prev_graph,
                        pred=None,
                    )
                    self.seqd[i + 1].query(verbose=verbose)
                    if len(self.seqd[i + 1].G) < 2:
                        break
                    self.log += self.seqd[i + 1].log
                    self.G = merge_two_networkx_graphs(self.G, self.seqd[i + 1].G)
                    self.output_ids[str(i + 1)] = self.seqd[i + 1].output_ids
                    self.prev_graph.update(self.seqd[i + 1].current_graph)
                    if i + 1 == len(self.paths):
                        self.current_graph.update(self.seqd[i + 1].current_graph)

        if verbose:
            print("\n==========")
            print("========== Final assembly of results ==========")
            print("==========\n\n")
            for i in range(len(self.paths)):
                if self.output_ids.get(str(i + 1)):
                    for output_cls, _ids in self.output_ids[str(i + 1)].items():
                        print(
                            "In the #{} query, BTE found {} unique {} nodes".format(
                                i + 1, len(_ids), output_cls
                            )
                        )
        self.log.append("\n==========")
        self.log.append("========== Final assembly of results ==========")
        self.log.append("==========\n\n")
        for i in range(len(self.paths)):
            if self.output_ids.get(str(i + 1)):
                for output_cls, _ids in self.output_ids[str(i + 1)].items():
                    self.log.append(
                        "In the #{} query, BTE found {} unique {} nodes".format(
                            i + 1, len(_ids), output_cls
                        )
                    )

    def to_json(self):
        """convert the graph into JSON through networkx"""
        if self.G.number_of_nodes() > 0:
            res = nx.json_graph.node_link_data(self.G)
            return res
        return {}

    def show_path(self, remove_duplicate=True):
        # if the last query was not performed, return
        if not self.output_ids.get(str(len(self.paths))):
            return
        # gather all outputs from the last query
        final_outputs = set()
        for output_ids in self.output_ids.get(str(len(self.paths))).values():
            for k, item in output_ids.items():
                if item.get("SYMBOL"):
                    final_outputs.add(item["SYMBOL"][0])
                elif item.get("name"):
                    final_outputs.add(item["name"][0])
                else:
                    splitted = k.split(":")
                    if len(splitted) == 2:
                        final_outputs.add(splitted[-1])
                    elif len(splitted) == 3:
                        final_outputs.add(":".join(splitted[1:]))
        # if the last query returns no results, return
        if len(final_outputs) == 0:
            return
        if len(self.paths) == 1:
            paths = []
            for _node in final_outputs:
                paths.append([self.starts, _node])
            return paths
        if remove_duplicate:
            paths = set()
            for _node in final_outputs:
                for path in nx.all_simple_paths(
                    self.G, source=self.starts, target=_node
                ):

                    path = "||".join(path)
                    paths.add(path)
                new_paths = []
                for _path in paths:
                    new_paths.append(_path.split("||"))
            return new_paths
        paths = []
        for _node in final_outputs:
            for path in nx.all_simple_paths(
                self.G, source=self.starts, target=self.ends
            ):
                paths.append(path)
        return paths

    def display_node_info(self, node):
        """show detailed node information

        Params
        ------
        node: str, node id
        """
        if node not in self.G:
            raise Exception("{} is not in the graph".format(node))
        return self.G.nodes[node]

    def display_edge_info(self, start_node, end_node):
        """display detailed edge info between start node and end node

        Params
        ------
        start_node: str, start node id
        end_node: str, end node id
        """
        if start_node not in self.G:
            raise Exception("{} is not in the graph".format(start_node))
        if end_node not in self.G:
            raise Exception("{} is not in the graph".format(end_node))
        if not self.G.has_edge(start_node, end_node):
            raise Exception(
                "No edge exists between {} and {}".format(start_node, end_node)
            )
        return dict(self.G[start_node][end_node])

    def show_all_nodes(self):
        """show all nodes in the graph"""
        return list(self.G.nodes())

    def show_all_edges(self):
        """show all edges in the graph"""
        return list(self.G.edges())

    def display_table_view(self):
        """Display the query results as a pandas table
        
        Examples
        --------
        >>> df = fc.display_table_view()
        >>> df
        """
        # paths = self.show_path()
        # return connect_networkx_to_pandas_df(self.G, paths)
        self.paths.insert(0, self.input_obj["type"])
        return networkx2pandas(self.current_graph, self.input_obj["type"])


class FindConnection:
    """Find relationships between one specific entity and another specific entity or other classes of entity types.

        Args:
            input_obj (required): must be an object returned from Hint corresponding to a specific biomedical entity.
                                Examples: 
                    Hint().query("Fanconi anemia")['DiseaseOrPhenotypicFeature'][0]
                    Hint().query("acetaminophen")['ChemicalSubstance'][0]

            output_obj (required): must EITHER be an object returned from Hint corresponding to a specific biomedical
                                entity, OR be a string or list of strings corresponding to Biolink Entity classes.
                                Examples:
                    Hint().query("acetaminophen")['ChemicalSubstance'][0]
                    'Gene'
                    ['Gene','ChemicalSubstance']

            intermediate_nodes (required): the semantic type(s) of the intermediate node(s).  Examples:
                    None                         : no intermediate node, find direct connections only
                    []                           : no intermediate node, find direct connections only
                    ['BiologicalEntity']         : one intermediate node of any semantic type
                    ['Gene']                     : one intermediate node that must be a Gene
                    [('Gene','Pathway')]         : one intermediate node that must be a Gene or a Pathway
                    ['Gene','Pathway']           : two intermediate nodes, first must be a Gene, second must be a Pathway.
                    ['Gene',('Pathway','Gene')]  : two intermediate nodes, first must be a Gene, second must be a Pathway or Gene.
                                                    **NOTE**: queries with more than one intermediate node are currently not supported
    
    **NOTE**: queries with more than one intermediate node are currently not supported
    """

    def __init__(self, input_obj, output_obj, intermediate_nodes, registry=None):
        """Find relationships in the Knowledge Graph between an Input Object and an Output Object.
        
        Args:
            input_obj (required): must be an object returned from Hint corresponding to a specific biomedical entity.
                                Examples:
                    Hint().query("Fanconi anemia")['DiseaseOrPhenotypicFeature'][0]
                    Hint().query("acetaminophen")['ChemicalSubstance'][0]

            output_obj (required): must EITHER be an object returned from Hint corresponding to a specific biomedical
                                entity, OR be a string or list of strings corresponding to Biolink Entity classes.
                                Examples:
                    Hint().query("acetaminophen")['ChemicalSubstance'][0]
                    'Gene'
                    ['Gene','ChemicalSubstance']

            intermediate_nodes (required): the semantic type(s) of the intermediate node(s).  Examples:
                    None                         : no intermediate node, find direct connections only
                    []                           : no intermediate node, find direct connections only
                    ['BiologicalEntity']         : one intermediate node of any semantic type
                    ['Gene']                     : one intermediate node that must be a Gene
                    [('Gene','Pathway')]         : one intermediate node that must be a Gene or a Pathway
                    ['Gene','Pathway']           : two intermediate nodes, first must be a Gene, second must be a Pathway.
                    ['Gene',('Pathway','Gene')]  : two intermediate nodes, first must be a Gene, second must be a Pathway or Gene.
                                                    **NOTE**: queries with more than one intermediate node are currently not supported
        """
        self.input_obj = input_obj
        self.output_obj = output_obj
        self.intermediate_nodes = intermediate_nodes
        if isinstance(output_obj, dict):
            self.fc = Explain(
                input_obj, output_obj, intermediate_nodes, registry=registry
            )
        else:
            self.fc = Predict(
                input_obj, output_obj, intermediate_nodes, registry=registry
            )

    def connect(self, verbose=False):
        self.fc.connect(verbose=verbose)

    def to_json(self):
        """Convert the graph into JSON through networkx."""
        return self.fc.to_json()

    def show_path(self, remove_duplicate=True):
        return self.fc.show_path(remove_duplicate=remove_duplicate)

    def display_node_info(self, node):
        """Show detailed node information.

        Parameters
            * node (str): node id
        """
        return self.fc.display_node_info(node)

    def display_edge_info(self, start_node, end_node):
        """Display detailed edge info between start node and end node.

        Parameters
            * start_node (str): start node id
            * end_node (str): end node id
        """
        return self.fc.display_edge_info(start_node, end_node)

    def show_all_nodes(self):
        """Show all nodes in the graph."""
        return self.fc.show_all_nodes()

    def show_all_edges(self):
        """Show all edges in the graph."""
        return self.fc.show_all_edges()

    def display_table_view(self):
        """Display the query results as a pandas table.
        
        **Examples**

        >>> df = fc.display_table_view()
        >>> df
        """
        return self.fc.display_table_view()

    def to_reasoner_std(self):
        """Convert the output to reasoner api standard.
        """
        rc = ReasonerConverter()
        rc.load_bte_query_path(
            start=self.input_obj,
            intermediate=self.intermediate_nodes,
            end=self.output_obj,
        )
        rc.load_bte_output(self.fc.G)
        return rc.generate_reasoner_response()

    def to_graphml(self, path):
        """Convert the output to graphml format.

        parameters
            * path (str): the file path to store the graphml file
        """
        rc = GraphmlConverter()
        rc.load_bte_output(self.fc.G)
        rc.generate_graphml_output(path)
        print("graphml file has been saved at {}".format(path))



from collections import defaultdict
from aiohttp import ClientTimeout
import traceback

from .parser import BioThingsParser

from ..config_new import (
    ALWAYS_PREFIXED,
    ID_RESOLVING_APIS,
    MAX_CONCURRENT_QUERIES_ON_SINGLE_API,
)


class Dispatcher:
    def __init__(self, inputIDs, session):
        self.inputIds = inputIDs
        self.tasks = defaultdict(list)
        self.invalid = {}
        self.session = session

    def dispatch(self):
        for semantic_type, curies in self.inputIds.items():
            res = self.generateQueryTasks(curies, semantic_type)
            for i in range(0, len(res["valid"]), MAX_CONCURRENT_QUERIES_ON_SINGLE_API):
                self.tasks[i] += res["valid"][
                    i : i + MAX_CONCURRENT_QUERIES_ON_SINGLE_API
                ]
            self.invalid[semantic_type] = res.get("invalid")

    def groupIdByPrefix(self, ids, semantic_type):
        if not ids:
            return {}
        mapping = {}
        res = defaultdict(set)
        for _id in ids:
            if isinstance(_id, int):
                res["invalid"].add(str(_id))
                continue
            if not isinstance(_id, str):
                continue
            splitted = _id.split(":", 1)
            if len(splitted) == 1:
                res["invalid"].add(_id)
            else:
                prefix = splitted[0]
                if semantic_type not in ID_RESOLVING_APIS:
                    res["invalid"].add(_id)
                    continue
                if prefix not in ID_RESOLVING_APIS[semantic_type]["mapping"]:
                    res["invalid"].add(_id)
                    continue
                val = splitted[-1]
                if prefix in ALWAYS_PREFIXED:
                    val = prefix + ":" + val
                mapping[prefix + ":" + val] = _id
                val = '"' + val + '"'
                res[prefix].add(val)
        res = dict(res)
        if mapping:
            res.update({"mapping": mapping})
        return res

    def generateQueryTasks(self, curies, semantic_type):
        inputs = self.groupIdByPrefix(curies, semantic_type)
        res = {"valid": [], "invalid": []}
        if not inputs:
            return res
        for prefix, ids in inputs.items():
            if prefix == "invalid":
                res["invalid"] = list(ids)
                continue
            if prefix == "mapping":
                res["mapping"] = ids
                continue
            ids = list(ids)
            for i in range(0, len(ids), 1000):
                query = self.constructQuery(ids[i : i + 1000], semantic_type, prefix)
                if not query:
                    res["invalid"] += [
                        (prefix + ":" + item) for item in ids[i : i + 1000]
                    ]
                else:
                    res["valid"].append(query)
        return res

    def constructQuery(self, inputs, semantic_type, prefix):
        """construct a BioThings batch query using axios
        The query aims to fetch all equivalent IDs for the inputs
        note: the input IDs must be less than 1000;
        The return value is an axios post query promises
        """
        query = "q={inputs}&scopes={scopes}&fields={fields}&dotfield=true&species=human"
        if not inputs:
            return
        if semantic_type not in ID_RESOLVING_APIS:
            return
        if prefix not in ID_RESOLVING_APIS[semantic_type]["mapping"]:
            return
        fields = []
        for item in ID_RESOLVING_APIS[semantic_type]["mapping"].values():
            fields += item
        scopes = ID_RESOLVING_APIS[semantic_type]["mapping"][prefix]
        query = (
            query.replace("{inputs}", ",".join(inputs))
            .replace("{scopes}", ",".join(scopes))
            .replace("{fields}", ",".join(fields))
        )
        return self.callSingleAPI(semantic_type, query, prefix)

    async def callSingleAPI(self, semanticType, query, prefix):
        async with self.session.post(
            ID_RESOLVING_APIS[semanticType]["url"] + "/query",
            data=query,
            timeout=ClientTimeout(20),
            headers={"content-type": "application/x-www-form-urlencoded"},
        ) as res:
            if res.status >= 400:

                print(
                    "api call to {} failed with status code {}".format(
                        ID_RESOLVING_APIS[semanticType]["url"] + "/query", res.status
                    )
                )
                print(res.json())
                return
            try:
                res = await res.json()
                parser = BioThingsParser(res, semanticType, prefix)
                res = parser.parse()
                return res
            except Exception as e:
                traceback.print_exc()
                print("failed to resolve {} ids".format(semanticType))
                return




class Print:
    def print_final_result_summary(self, stepsNodes):
        """
        Print the summary of final results.
        :param: stepsNodes: Unique node objects retrieved from each query step
        """
        print(
            "\n==========\n========== Final assembly of results ==========\n==========\n"
        )
        for k, j in stepsNodes.items():
            print("In the #{} query, BTE found {} unique nodes.".format(k + 1, len(j)))

    def print_query_failure_message(self, reason, query_id):
        """
        Print message regarding why the query fails.
        :param reason: Reason for query failure
        """
        print(
            "Unfortunately, no {} were found for Query # {}. Your query fails. Please try refine your query!".format(
                reason, query_id
            )
        )

    def print_expand_begin_message(self):
        print("\n==========\n========== QUERY EXPANSION ==========\n==========\n")
        print("You have selected to expand your inputs.")

    def print_query_plan_begin_message(self, query_id):
        print("\n========== Query # {} ==========\n".format(query_id))
        print("==== Step #1: Query Path Planning ====\n")

    def print_expand_summary_message(self, expandedInputs):
        """
        Print a summary of the expansion results.
        :param expandedInputs: A list of results from expansion.
        """
        if expandedInputs:
            print(
                "{} unique bioentities have been found as subclass of your inputs.\n".format(
                    len(expandedInputs)
                )
            )
        else:
            print("No bioentity was found to be subclass of your inputs.\n")

    def print_query_parameter_summary(self, inputs, intermediate_nodes):
        """
        Print a summary of query parameters.
        :param inputs: list of input nodes.
        :param intermediate_nodes: list of intermediate nodes.
        """
        print("==========\n========== QUERY PARAMETER SUMMARY ==========\n==========\n")
        print(
            "Your query have {} input nodes, including {} .... And BTE will find paths that connect your input nodes to your output types {}. \
Paths will contain {} intermediate nodes.\n".format(
                len(inputs),
                ",".join([item["id"]["label"] for item in inputs.values()][0:5]),
                intermediate_nodes[-1],
                len(intermediate_nodes) - 1,
            )
        )
        for i, node in enumerate(intermediate_nodes[0:-1]):
            print(
                "Intermediate node #{} will have these type constraints: {}".format(
                    i + 1, node
                )
            )

    def print_individual_query_step_summary(self, outputs):
        """
        Print summary of each query steps.
        :param outputs: outputs from the query step.
        """
        print(
            "After id-to-object translation, BTE retrieved {} unique output nodes.\n\n".format(
                len(outputs)
            )
        )

